#include "PreCompiledHeaders.hh"
#include "Lib/Ass/Style.hh"
#include "Lib/Utils.hh"

using namespace Vivy::Ass;

Style::Style(const QString &styleString)
{
    // Some usefull defines only needed in that function

    enum StyleIndex : int {
        Name     = 0,
        FontName = 1,
        FontSize = 2,

        ColorPrimary   = 3,
        ColorSecondary = 4,
        ColorOutline   = 5,
        ColorBack      = 6,

        Bold      = 7,
        Italic    = 8,
        Underline = 9,
        StrikeOut = 10,

        ScaleX  = 11,
        ScaleY  = 12,
        Spacing = 13,
        Angle   = 14,

        BorderStyle = 15,

        Outline = 16,
        Shadow  = 17,

        Alignement = 18,
        MarginL    = 19,
        MarginR    = 20,
        MarginV    = 21,

        Encoding = 22,

        PastLastCode
    };

    const QString lineHeader = QStringLiteral("Style: ");

    // Check line header and content items number

    if (!styleString.startsWith(lineHeader))
        throw std::runtime_error(("invalid style line header: " + styleString).toStdString());

    const QString lineContent = styleString.mid(styleString.indexOf(": ") + 2 /* strlen ": " */);
    const QStringList content = lineContent.split(",", Qt::KeepEmptyParts, Qt::CaseInsensitive);

    if (content.size() != StyleIndex::PastLastCode)
        throw std::runtime_error(("invalid number of items " + QString::number(content.size()) +
                                  " instead of " + QString::number(PastLastCode) +
                                  " for line: " + lineContent)
                                     .toStdString());

    // Unpack data from the line

    styleName = content[StyleIndex::Name];
    fontName  = content[StyleIndex::FontName];
    fontSize =
        Utils::decodeLineToInteger(content[StyleIndex::FontSize], "FontSize is not an integer");

    primaryColor   = Color::fromString(content[StyleIndex::ColorPrimary]);
    secondaryColor = Color::fromString(content[StyleIndex::ColorSecondary]);
    outlineColor   = Color::fromString(content[StyleIndex::ColorOutline]);
    backColor      = Color::fromString(content[StyleIndex::ColorBack]);

    bold   = Utils::decodeLineToBoolean(content[StyleIndex::Bold], "Bold is not a boolean");
    italic = Utils::decodeLineToBoolean(content[StyleIndex::Italic], "Italic is not a boolean");
    underline =
        Utils::decodeLineToBoolean(content[StyleIndex::Underline], "Underline is not a boolean");
    strikeOut =
        Utils::decodeLineToBoolean(content[StyleIndex::StrikeOut], "StrikeOut is not a boolean");

    scaleX  = Utils::decodeLineToFloating(content[StyleIndex::ScaleX], "ScaleX is not a floating");
    scaleY  = Utils::decodeLineToFloating(content[StyleIndex::ScaleY], "ScaleY is not a floating");
    spacing = Utils::decodeLineToFloating(content[StyleIndex::ScaleY], "Spacing is not a floating");
    angle   = Utils::decodeLineToFloating(content[StyleIndex::Angle], "Angle is not a floating");
    borderStyle = Utils::decodeLineToFloating(content[StyleIndex::BorderStyle],
                                              "BorderStyle is not a floating");
    outline =
        Utils::decodeLineToFloating(content[StyleIndex::Outline], "Outline is not a floating");
    shadow = Utils::decodeLineToFloating(content[StyleIndex::Shadow], "Shadow is not a floating");

    alignment =
        Utils::decodeLineToInteger(content[StyleIndex::Alignement], "Alignement is not an integer");
    if (alignment < 1 || alignment > 9)
        throw std::runtime_error(
            ("invalid line content: Alignement must be between 1 and 9 included, it was " +
             QString::number(alignment))
                .toStdString());

    marginL = Utils::decodeLineToInteger(content[StyleIndex::MarginL], "MarginL is not an integer");
    marginR = Utils::decodeLineToInteger(content[StyleIndex::MarginR], "MarginR is not an integer");
    marginV = Utils::decodeLineToInteger(content[StyleIndex::MarginV], "MarginV is not an integer");
    encoding =
        Utils::decodeLineToInteger(content[StyleIndex::Encoding], "Encoding is not an integer");

    if (encoding != 1)
        logWarning() << "Encoding is not '1' in the ASS Style";
}

QString
Color::toString(const QColor &c) noexcept
{
    return "&H" + QString::number(c.blue(), 16).toUpper() +
           QString::number(c.green(), 16).toUpper() + QString::number(c.red(), 16).toUpper();
}

QColor
Color::fromString(const QString &colorString) noexcept
{
    // Ignore the '&H' at the begeing of the color if needed
    int startIndex = 0;
    if (colorString[startIndex] == '&')
        startIndex++;
    if (colorString[startIndex] == 'H' || colorString[startIndex] == 'h')
        startIndex++;

    // In some cases, the color can begin by a '#'
    if (colorString[startIndex] == '#')
        startIndex++;

    // A valid string color is like 'AARRGGBB' for now (skipped 'aH')
    if (colorString.size() - startIndex != 8)
        return Color::defaultValue;

    bool ok_alpha = false;
    bool ok_red   = false;
    bool ok_green = false;
    bool ok_blue  = false;
    int alpha     = colorString.mid(startIndex, 2).toInt(&ok_alpha, 16);
    int red       = colorString.mid(startIndex + 2, 2).toInt(&ok_red, 16);
    int green     = colorString.mid(startIndex + 4, 2).toInt(&ok_green, 16);
    int blue      = colorString.mid(startIndex + 6, 2).toInt(&ok_blue, 16);

    if (!(ok_alpha && ok_red && ok_green && ok_blue))
        return Color::defaultValue;

    return QColor(red, green, blue, alpha);
}

QString
Style::getElementName() const noexcept
{
    return styleName;
}

QJsonDocument
Style::getProperties() const noexcept
{
    QJsonDocument ret;

    QJsonObject styleFont{
        { "Font name", fontName }, { "Font size", fontSize },  { "Bold", bold },
        { "Italic", italic },      { "Underline", underline }, { "Strike Out", strikeOut },
    };

    QJsonObject styleColors{
        { "Primary", Color::toString(primaryColor) },
        { "Secondary", Color::toString(secondaryColor) },
        { "Outline", Color::toString(outlineColor) },
        { "Back", Color::toString(backColor) },
    };

    QJsonObject styleFontProperties{
        { "Scale X", static_cast<double>(scaleX) },
        { "Scale Y", static_cast<double>(scaleY) },
        { "Spacing", static_cast<double>(spacing) },
        { "Angle", static_cast<double>(angle) },
        { "Border Style", static_cast<double>(borderStyle) },
        { "Outline", static_cast<double>(outline) },
        { "Shadow", static_cast<double>(shadow) },
    };

    QJsonObject styleMargins{
        { "Left", marginL },
        { "Right", marginR },
        { "Vertical", marginV },
    };

    QJsonObject object{ { "Name", styleName },
                        { "Font", styleFont },
                        { "Font properties", styleFontProperties },
                        { "Colors", styleColors },
                        { "Margin", styleMargins },
                        { "Alignement", alignment },
                        { "Encoding", encoding } };

    ret.setObject(object);
    return ret;
}

#define PRINT_STYLE_BOOL(b) b ? -1 : 0
QString
Style::getRawText() const noexcept
{
    // FIXME: asprintf bad, use QString("%x").arg()?
    return QString::asprintf(
        "Style: %s,%s,%d,%s,%s,%s,%s,%d,%d,%d,%d,%0.2f,%0.2f,%0.2f,%0.2f,%0.2f,%0.2f,%0.2f,%d,%d,%d,%d,%d",
        styleName.toStdString().c_str(), fontName.toStdString().c_str(), fontSize,
        Color::toString(primaryColor).toStdString().c_str(),
        Color::toString(secondaryColor).toStdString().c_str(),
        Color::toString(outlineColor).toStdString().c_str(),
        Color::toString(backColor).toStdString().c_str(), PRINT_STYLE_BOOL(bold),
        PRINT_STYLE_BOOL(italic), PRINT_STYLE_BOOL(underline), PRINT_STYLE_BOOL(strikeOut),
        double(scaleX), double(scaleY), double(spacing), double(angle), double(borderStyle),
        double(outline), double(shadow), alignment, marginL, marginR, marginV, encoding);
}
#undef PRINT_STYLE_BOOL
