#include "PreCompiledHeaders.hh"
#include "Lib/Document/VivyDocument.hh"
#include "Lib/Utils.hh"
#include "Lib/JsonBuilder.hh"

using namespace Vivy;

// Runtime document version, don't set it in the header!!!
QString VivyDocument::runtimeVersion = QStringLiteral("alpha");

VivyDocument::VivyDocument(const QString &documentPath, Options opt)
    : AbstractDocument(AbstractDocument::Type::Vivy, documentPath)
    , documentOptions(opt)
{
    // Ignore some things in case of memory document creation because no real
    // document with an extension is available.
    if (opt & MemoryDocumentCreation) {
        documentName     = documentPath;
        documentLocation = QDir::home();
    }

    // Load a real document
    else {
        QFileInfo fileInfo(name);
        QFile file(name);
        documentName     = fileInfo.baseName();
        documentLocation = fileInfo.absoluteDir();

        if ((fileInfo.completeSuffix() != fileSuffix))
            throw std::runtime_error("Invalid file suffix");

        if (!file.exists() || !fileInfo.isFile() ||
            !file.open(QIODevice::ReadOnly | QIODevice::Text))
            throw std::runtime_error("Document doesn't exists");

        QTextStream fileContent(&file);
        QJsonParseError jsonError;
        QJsonDocument json = QJsonDocument::fromJson(fileContent.readAll().toUtf8(), &jsonError);
        if (json.isNull())
            throw std::runtime_error("Json error " + jsonError.errorString().toStdString());

        loadSaveJsonDocumentFile(json);
    }
}

void
VivyDocument::save()
{
    if (documentOptions & MemoryDocumentCreation)
        throw std::logic_error("You can't save a memory document");

    const bool needExt     = !documentName.endsWith(fileSuffix);
    const QString filepath = documentLocation.absolutePath() + QDir::separator() + documentName +
                             (needExt ? "." + fileSuffix : "");
    QFile file(filepath);

    if (!file.open(QIODevice::WriteOnly | QIODevice::Text | QIODevice::Truncate))
        throw std::runtime_error("Failed to open file in WO+Text mode: " + filepath.toStdString());

    QTextStream writter(&file);
    QJsonDocument json = getSaveJsonDocumentFile();
    writter << json.toJson(QJsonDocument::Compact);
    file.close();
    documentOptions = static_cast<Options>(documentOptions & (~UntouchedByDefault));
}

void
VivyDocument::loadSaveJsonDocumentFile_ALPHA(VivyDocument *const self, const QJsonDocument &json)
{
    // Init from the JSON
    self->documentType = json[KeyCapabilities].toString().toULongLong() & possibleCapabilities;
    if (json[KeyName].toString() != self->documentName)
        throw std::runtime_error("The document was edited outside of Vivy");

    if (QJsonValue audio = json[KeySubDocuments][KeySubAudio];
        !audio.isUndefined() && !self->loadSubDocument(audio.toString(), Capabilities::AudioAble))
        throw std::runtime_error("Failed to load audio sub document");

    if (QJsonValue video = json[KeySubDocuments][KeySubVideo];
        !video.isUndefined() && !self->loadSubDocument(video.toString(), Capabilities::VideoAble))
        throw std::runtime_error("Failed to load video sub document");

    /*
     * TODO: for now, if there is an internal ass we use it and "ignore" the SubAss file
     * In the future, both should be loaded to allow the user to switch between the 2
     * (e.g. a "reference" ass and the vivy ass)
     */
    if (QJsonValue haveInternalAss = json[KeySubDocuments][KeyHaveInternalAss];
        !haveInternalAss.isUndefined() && haveInternalAss.toBool()) {
        // ASS is inside Vivy document
        if (QJsonValue internalAssSource = json[KeySubDocuments][KeyInternalAssSource];
            !internalAssSource.isUndefined() && internalAssSource.isArray()) {
            if (!self->loadSubDocument(internalAssSource,
                                       json[KeySubDocuments][KeySubAss].toString(),
                                       Capabilities::AssAble)) {
                throw std::runtime_error("Failed to load ASS sub document");
            }
        }
    } else if (QJsonValue ass = json[KeySubDocuments][KeySubAss]; !ass.isUndefined()) {
        // ASS in its own ASS file
        if (!self->loadSubDocument(ass.toString(), Capabilities::AssAble))
            throw std::runtime_error("Failed to load ASS sub document");
    }
}

void
VivyDocument::loadSaveJsonDocumentFile(const QJsonDocument &json)
{
    if (json.isNull())
        throw std::logic_error("Json is null, you can't pass a null json "
                               "document to this function");

    const QJsonValue savedVersionJson = json[KeyDocumentVersion];
    if (savedVersionJson.isUndefined())
        throw std::runtime_error("No version present in the saved VivyDocument");

    const QString savedVersion = savedVersionJson.toString();
    for (const auto &[version, callback] : loadFunctions) {
        if (savedVersion == QString::fromUtf8(version))
            return callback(this, json);
    }

    throw std::runtime_error("Version is undefined! No way has been "
                             "found to load the document");
}

QJsonDocument
VivyDocument::getSaveJsonDocumentFile() const
{
    QJsonDocument ret;
    QJsonObject json;
    QJsonObject subDocumentJson;

    json.insert(KeyCapabilities, QString::number(possibleCapabilities & documentType));
    json.insert(KeyName, documentName);
    json.insert(KeyDocumentVersion, runtimeVersion);

    if (documentType & Capabilities::AudioAble)
        subDocumentJson.insert(KeySubAudio, audioDocument->getFilePath());

    if (documentType & Capabilities::VideoAble)
        subDocumentJson.insert(KeySubVideo, videoDocument->getFilePath());

    if (documentType & Capabilities::AssAble) {
        subDocumentJson.insert(KeySubAss, assDocument->getFilePath());
        subDocumentJson.insert(KeyHaveInternalAss, assDocument->haveInternalAss());

        // ASS is inside Vivy document
        if (assDocument->haveInternalAss())
            subDocumentJson.insert(KeyInternalAssSource, assDocument->getInternalAss());
    }

    json.insert(KeySubDocuments, subDocumentJson);
    ret.setObject(json);
    return ret;
}

bool
VivyDocument::loadSubDocument(const QString &subName) noexcept
{
    QFileInfo file(subName);
    Capabilities ableType;

    if (detectDocumentType(file, &ableType)) {
        switch (ableType) {
        case Capabilities::AudioAble:
            logDebug() << "Auto-detect audio document for " << subName;
            setAudioSubDocument(file.absoluteFilePath());
            return true;

        case Capabilities::VideoAble:
            logDebug() << "Auto-detect video (and try audio) document for " << subName;
            setVideoSubDocument(file.absoluteFilePath());
            setAudioSubDocument(file.absoluteFilePath());
            return true;

        case Capabilities::AssAble:
            logDebug() << "Auto-detect ASS document for " << subName;
            setAssSubDocument(file.absoluteFilePath());
            return true;
        }
    }

    return false;
}

bool
VivyDocument::loadSubDocument(const QString &subName, VivyDocument::Capabilities asType) noexcept
{
    QFileInfo file(subName);
    Capabilities ableType;
    if (!detectDocumentType(file, &ableType)) {
        logError() << "Failed to detect type for file " << subName;
        return false;
    }

    if (ableType == Capabilities::AssAble && asType == Capabilities::AssAble) {
        logDebug() << "Create an ass subDocument from " << subName;
        setAssSubDocument(file.absoluteFilePath());
    }

    else if (ableType == Capabilities::VideoAble && asType == Capabilities::VideoAble) {
        logDebug() << "Create a video subDocument from " << subName;
        setVideoSubDocument(file.absoluteFilePath());
    }

    else if (const bool requestAudio = (asType == Capabilities::AudioAble);
             (ableType == Capabilities::VideoAble && requestAudio) ||
             (ableType == Capabilities::AudioAble && requestAudio)) {
        logDebug() << "Create an audio subDocument from " << subName;
        setAudioSubDocument(file.absoluteFilePath());
    }

    return true;
}

bool
VivyDocument::loadSubDocument(const QJsonValue &internalAss, const QString &baseAss,
                              VivyDocument::Capabilities asType) noexcept
{
    logDebug() << "ASS sub-doc: Trying to open file internal ASS";

    if (asType == Capabilities::AssAble) {
        if ([[maybe_unused]] auto doc = assDocument.get()) {
            // Here we may want to do some confirmation / cleanup
            assDocument.reset();
        }

        assDocument = AssSubDocument::fromInternal(internalAss, baseAss);
        if (assDocument)
            addDocumentType(AssAble);
    }

    return true;
}

bool
VivyDocument::detectDocumentType(const QFileInfo &file, Capabilities *const ableType) noexcept
{
    Vivy::Utils::DocumentType docType;
    bool rc = Vivy::Utils::detectDocumentType(file, &docType);

    if (docType == Utils::DocumentType::Video)
        *ableType = Capabilities::VideoAble;

    else if (docType == Utils::DocumentType::Audio)
        *ableType = Capabilities::AudioAble;

    else if (docType == Utils::DocumentType::ASS)
        *ableType = Capabilities::AssAble;

    // Invalid document type
    else
        return false;

    return rc;
}

void
VivyDocument::copy(const QString &newName)
{
    const QFileInfo newPath = computeFileInfo(newName);

    // Rename + create the disk on disk as it's a memory document
    if (documentOptions & MemoryDocumentCreation)
        saveMemoryFile(newPath);

    // Compute new paths, the document is really on disk initially
    else {
        logDebug() << "Renaming a real file to " << newName;
        copyWith(newPath, [=, this]() noexcept -> void {
            documentLocation = newPath.dir();
            documentName     = newPath.baseName();
            name             = newPath.absoluteFilePath();
        });
    }

    emit documentChanged();
}

void
VivyDocument::rename(const QString &newName)
{
    const QFileInfo newPath = computeFileInfo(newName);

    // Rename + create the disk on disk as it's a memory document
    if (documentOptions & MemoryDocumentCreation)
        saveMemoryFile(newPath);

    // Compute new paths, the document is really on disk initially
    else {
        logDebug() << "Renaming a real file to " << newName;
        renameWith(newPath, [=, this]() noexcept -> void {
            documentLocation = newPath.dir();
            documentName     = newPath.baseName();
            name             = newPath.absoluteFilePath();
        });
    }

    emit documentChanged();
}

QFileInfo
VivyDocument::computeFileInfo(const QString &newName) const noexcept
{
    const QString newNameWithExtension =
        (newName.right(fileSuffix.size()) == fileSuffix) ? newName : newName + "." + fileSuffix;
    return QFileInfo(documentLocation, newNameWithExtension);
}

void
VivyDocument::saveMemoryFile(const QFileInfo &newPath)
{
    documentName     = newPath.baseName();
    documentLocation = newPath.absoluteDir();
    name             = newPath.absoluteFilePath();
    documentOptions  = static_cast<Options>(documentOptions & (~MemoryDocumentCreation));
    logDebug() << "Renaming a memory file => create it on disk with { "
               << documentLocation.absolutePath() << ", " << documentName << " }";
    save();
}

std::shared_ptr<AudioSubDocument>
VivyDocument::getAudioSubDocument() const noexcept
{
    return audioDocument;
}

std::shared_ptr<VideoSubDocument>
VivyDocument::getVideoSubDocument() const noexcept
{
    return videoDocument;
}

std::shared_ptr<AssSubDocument>
VivyDocument::getAssSubDocument() const noexcept
{
    return assDocument;
}

AudioContext::StreamPtr
VivyDocument::getAudioStream() const noexcept
{
    if (auto p = audioDocument.get())
        return audioDocument->getDefaultStream();
    return AudioContext::StreamPtr();
}

bool
VivyDocument::checkDocumentCapabilities(Capabilities asType) const noexcept
{
    return static_cast<bool>(documentType & asType);
}

bool
VivyDocument::checkDocumentOption(Options opt) const noexcept
{
    return static_cast<bool>(documentOptions & opt);
}

QString
VivyDocument::getDocumentCapabilitiesString() const noexcept
{
    QStringList ret;
    if (documentType & AudioAble)
        ret.push_back("AudioAble" + QString::fromUtf8(audioDocument ? "" : "(ack)"));
    if (documentType & VideoAble)
        ret.push_back("VideoAble" + QString::fromUtf8(videoDocument ? "" : "(ack)"));
    if (documentType & AssAble)
        ret.push_back("AssAble" + QString::fromUtf8(assDocument ? "" : "(ack)"));
    return ret.join(", ");
}

void
VivyDocument::setAudioSubDocument(const QString filename) noexcept
{
    logDebug() << "Audio sub-doc: Trying to open file " << filename;
    QFileInfo fileInfo(filename);
    const QString baseName = fileInfo.baseName();

    if ([[maybe_unused]] auto doc = audioDocument.get()) {
        // Here we may want to do some confirmation / cleanup
        audioDocument.reset();
    }

    audioDocument = AudioSubDocument::fromFile(filename);
    if (audioDocument)
        addDocumentType(AudioAble);
}

void
VivyDocument::setVideoSubDocument(const QString filename) noexcept
{
    logDebug() << "Video sub-doc: Trying to open file " << filename;
    QFileInfo fileInfo(filename);
    const QString baseName = fileInfo.baseName();

    if ([[maybe_unused]] auto doc = videoDocument.get()) {
        // Here we may want to do some confirmation / cleanup
        videoDocument.reset();
    }

    videoDocument = VideoSubDocument::fromFile(filename);
    if (videoDocument)
        addDocumentType(VideoAble);
}

void
VivyDocument::setAssSubDocument(const QString filename) noexcept
{
    logDebug() << "ASS sub-doc: Trying to open file " << filename;
    QFileInfo fileInfo(filename);
    const QString baseName = fileInfo.baseName();

    if ([[maybe_unused]] auto doc = assDocument.get()) {
        // Here we may want to do some confirmation / cleanup
        assDocument.reset();
    }

    assDocument = AssSubDocument::fromFile(filename);
    if (assDocument)
        addDocumentType(AssAble);
}

void
VivyDocument::addDocumentType(Capabilities newCap) noexcept
{
    documentType |= newCap;
    documentOptions &= ~Options::UntouchedByDefault;
    emit documentChanged();
}

QString
VivyDocument::getElementName() const noexcept
{
    return getName();
}

QJsonDocument
VivyDocument::getProperties() const noexcept
{
    return JsonBuilder::createOrderedJsonDocument(
        { { QStringLiteral("Uuid"), getUuid().toString() },
          { QStringLiteral("Name"), getName() },
          { QStringLiteral("Sub documents"),
            JsonBuilder::createOrderedJsonObject({
                JsonBuilder::getSubDocumentJson("Audio sub document", audioDocument),
                JsonBuilder::getSubDocumentJson("Video sub document", videoDocument),
                JsonBuilder::getSubDocumentJson("ASS sub document", assDocument),
            }) } });
}
