//! Read the content of an ASS file / a Vivy subtitle file and creates an
//! [vvs_ass::elements::lines::ASSLinesPtr] structure accordingly.

use crate::{ASSContainer, AuxTable};
use std::{
    fs::File,
    io::{self, BufReader, Error as IoError},
    path::{Path, PathBuf},
};
use thiserror::Error;
use vvs_procmacro::EnumVariantFromStr;

mod ass;
mod json;

#[derive(Debug, EnumVariantFromStr)]
pub enum ContainerFileType {
    ASS,
    VVSB,
    Json,
}

#[derive(Debug, Error)]
pub enum ASSElementReaderError {
    #[error("file has no extension: {0}")]
    NoExtension(PathBuf),

    #[error("failed to open file {0}: {1}")]
    FailedToOpenFile(PathBuf, IoError),

    #[error("unknown file extension for subtitles")]
    UnknownExtension(String),

    #[error("failed to read line: {0}")]
    FailedToReadLine(IoError),

    #[error("{0}")]
    Custom(String),
}

trait ASSElementReader<T: AuxTable> {
    fn try_read(self, file: impl io::BufRead) -> Result<ASSContainer<T>, ASSElementReaderError>;
}

pub fn ass_container_from_file<T: AuxTable>(file: impl AsRef<Path>) -> Result<ASSContainer<T>, ASSElementReaderError> {
    let file = file.as_ref();
    let Some(extension) = file.extension() else {
        return Err(ASSElementReaderError::NoExtension(file.to_path_buf()));
    };
    let content = BufReader::new(
        File::open(file).map_err(|err| ASSElementReaderError::FailedToOpenFile(file.to_path_buf(), err))?,
    );
    match &extension.to_string_lossy().to_lowercase()[..] {
        "ass" => ass::ASSReader::default().try_read(content),
        "vvsb" | "json" => json::JSONReader::default().try_read(content),
        extension => Err(ASSElementReaderError::UnknownExtension(extension.to_string())),
    }
}

pub fn ass_container_from_str<T: AuxTable>(
    extension: ContainerFileType,
    str: impl AsRef<str>,
) -> Result<ASSContainer<T>, ASSElementReaderError> {
    use ContainerFileType::*;
    let content = BufReader::new(str.as_ref().as_bytes());
    match extension {
        ASS => ass::ASSReader::default().try_read(content),
        VVSB | Json => json::JSONReader::default().try_read(content),
    }
}
