use crate::{ASSAlign, ASSColor};

#[derive(Debug, Default, Clone, Copy, PartialEq, Eq)]
pub enum ASSBorderStyle {
    #[default]
    OutlineAndDropShadow = 1,
    OpaqueBox = 3,
}

/// Type used to describe an ASS style.
#[derive(Debug, Clone, PartialEq)]
pub struct ASSStyle {
    /// The name of the Style. Case sensitive. Cannot include commas.
    pub name: String,

    /// The fontname as used by libass. Case-sensitive. Can't include comas.
    pub font_name: String,

    /// The size of the font. Don't use a thing that is too big... Must be positive.
    pub font_size: i32,

    /// The colour that a subtitle will normally appear in.
    pub primary_color: ASSColor,

    /// This colour may be used instead of the Primary colour when a subtitle is automatically
    /// shifted to prevent an onscreen collsion, to distinguish the different subtitles.
    pub secondary_color: ASSColor,

    /// Color of the outline of each characters.
    pub outline_color: ASSColor,

    /// This is the colour of the subtitle outline or shadow, if these are used.
    pub back_color: ASSColor,

    /// Whever the font is in bold.
    pub bold: bool,

    /// Whever the font is in italic.
    pub italic: bool,

    /// Whever the font is underlined.
    pub underline: bool,

    /// Whever the font is strikeout.
    pub strikeout: bool,

    /// Modifies the width of the font. Value in percent.
    pub scale_x: f64,

    /// Modifies the height of the font. Value in percent.
    pub scale_y: f64,

    /// Extra space between characters. Value in pixels.
    pub spacing: f64,

    /// The origin of the rotation is defined by the alignment. Can be a floating point number.
    /// Value in degrees.
    pub angle: f64,

    /// The border style.
    pub border_style: ASSBorderStyle,

    /// If border_style is [ASSBorderStyle::OutlineAndDropShadow], then this specifies the width of
    /// the outline around the text, in pixels.
    ///
    /// Values may be 0, 1, 2, 3 or 4 the documentation says, but libass seems to support floating
    /// values.
    pub outline: f64,

    /// If border_style is [ASSBorderStyle::OutlineAndDropShadow], then this specifies the depth of
    /// the drop shadow behind the text, in pixels.
    ///
    /// Values may be 0, 1, 2, 3 or 4 the documentation says, but libass seems to support floating
    /// values. Drop shadow is always used in addition to an outline - SSA will force an outline of
    /// 1 pixel if no outline width is given.
    pub shadow: f64,

    /// This sets how text is "justified" within the Left/Right onscreen margins, and also the
    /// vertical placing.
    pub alignment: ASSAlign,

    /// This defines the Left Margin in pixels. It is the distance from the left-hand edge of the
    /// screen.The three onscreen margins (margin_l, margin_r, margin_v) define areas in which the
    /// subtitle text will be displayed.
    ///
    /// Must be a positive integer.
    pub margin_l: i32,

    /// This defines the Right Margin in pixels. It is the distance from the right-hand edge of the
    /// screen. The three onscreen margins (margin_l, margin_r, margin_v) define areas in which the
    /// subtitle text will be displayed.
    pub margin_r: i32,

    /// This defines the vertical Left Margin in pixels.
    /// - For a subtitle, it is the distance from the bottom of the screen.
    /// - For a toptitle, it is the distance from the top of the screen.
    /// - For a midtitle, the value is ignored - the text will be vertically centred
    pub margin_v: i32,
}

impl std::str::FromStr for ASSBorderStyle {
    type Err = String;

    #[inline]
    fn from_str(s: &str) -> Result<Self, Self::Err> {
        match s.trim() {
            "1" => Ok(ASSBorderStyle::OutlineAndDropShadow),
            "3" => Ok(ASSBorderStyle::OpaqueBox),
            s => Err(format!("invalid value '{s}' for ASSBorderStyle, must be 1 or 3")),
        }
    }
}

impl ASSStyle {
    #[inline]
    pub fn default_name() -> &'static str {
        "Default"
    }

    #[inline]
    pub fn default_name_string() -> String {
        Self::default_name().to_string()
    }
}

impl Default for ASSStyle {
    fn default() -> Self {
        let font = vvs_font::embeded_fonts()
            .next()
            .expect("we should at least bundle one font");
        let font = vvs_font::Font::try_from(font.1).expect("the font we bundle is not valid");
        let font_name = font.name().expect("the font we bundle doesn't have a valid name");
        match &font.font_types()[..] {
            [vvs_font::FontType::Regular] => {}
            _ => panic!("the font we bundle is not regular"),
        }

        Self {
            name: Self::default_name_string(),
            font_name,
            font_size: 14,
            primary_color: ASSColor::WHITE,
            secondary_color: ASSColor::RED,
            outline_color: ASSColor::BLACK,
            back_color: ASSColor::BLACK,
            bold: false,
            italic: false,
            underline: false,
            strikeout: false,
            scale_x: 100.0,
            scale_y: 100.0,
            spacing: 0.0,
            angle: 0.0,
            border_style: Default::default(),
            outline: 3.0,
            shadow: 1.0,
            alignment: ASSAlign::TC,
            margin_l: 0,
            margin_r: 0,
            margin_v: 10,
        }
    }
}
