use crate::parser::FileTypeValueParser;
use clap_complete::Shell;
use std::path::PathBuf;

#[derive(clap::Parser, Debug)]
#[command( author
         , version
         , about
         , name   = "vvcc"
         , groups = [ clap::ArgGroup::new("action").args(["manpage", "shell", "font-file", "script.vvs"])
                    , clap::ArgGroup::new("ass")   .args(["subtitle.ass"]).conflicts_with_all(["manpage", "font-file"])
                    , clap::ArgGroup::new("opts")  .args(["options.ini"]) .conflicts_with_all(["shell", "manpage", "font-file"])
])]
pub struct Args {
    /// The script to run.
    ///
    /// A script must not touch to fields that begins by underscores, never. Those fields are
    /// reserved for vivy or for library developpers. Note that even library developpers must not
    /// touch or call fields from vivy that begins by underscores.
    #[arg( action       = clap::ArgAction::Set
         , id           = "script.vvs"
         , value_parser = FileTypeValueParser::new(vvs_lang::VIVY_SCRIPT_EXTENSION)
    )]
    pub script: Option<PathBuf>,

    /// The input ASS file to execute the script on.
    ///
    /// The supported subtitle extensions are the ASS files (V4+) or the Json files exported by the
    /// Vivy application.
    #[arg ( short        = 'f'
          , long         = "subtitle"
          , action       = clap::ArgAction::Set
          , id           = "subtitle.ass"
          , value_parser = FileTypeValueParser::new("ass"),
    )]
    pub ass_file: Option<PathBuf>,

    /// The option file that will be used when running the script.
    #[arg( short        = 't'
         , long         = "option"
         , action       = clap::ArgAction::Set
         , id           = "options.ini"
         , value_parser = FileTypeValueParser::new("ini")
    )]
    pub options: Option<PathBuf>,

    /// The includes' folder list, in order.
    ///
    /// Will search for modules to import in those folders. Note that there is no need to add the
    /// folder of the loaded script as it will automatically be added in the top position of the
    /// include list.
    #[arg( short  = 'I'
         , long   = "include-path"
         , action = clap::ArgAction::Append
         , id     = "include"
    )]
    pub includes: Vec<PathBuf>,

    /// Display infos about embeded fonts or external fonts.
    #[arg( long   = "font-info"
         , action = clap::ArgAction::Set
         , id     = "font-file"
    )]
    pub font_info: Option<Option<PathBuf>>,

    /// Make vvcc more verbose, repeat to make it even more verbose
    #[arg( long
         , short  = 'v'
         , action = clap::ArgAction::Count
    )]
    pub verbose: u8,

    /// Generate the manpage for vvcc and print it out to stdout
    ///
    /// You may read it by running `vvcc --manpage | man -l -` or writing it in the
    /// `$HOME/.local/share/man/man1/` or `/usr/share/man/man1` folder
    #[arg( long   = "manpage"
         , action = clap::ArgAction::SetTrue
         , hide   = true
    )]
    pub manpage: bool,

    /// Generate completion script for the given shell
    #[arg( long   = "shell"
         , action = clap::ArgAction::Set
         , hide   = true
    )]
    pub shell: Option<Shell>,
}
