//! Contains structs to lower from Lua code to LLVM-IR

mod types;

use crate::{lowerer::types::TypeLowerer, CodegenError};
use vvs_lang::MainProgram;
use vvs_llvm::prelude::*;
use vvs_parser::prelude::ast;
use vvs_runtime_types::{VVRTSymbol, VVRTSymbolType};

pub struct Lowerer<'ctx, 'this>
where
    'ctx: 'this,
{
    context: &'this Context<'ctx>,
    module: Module<'ctx>,
}

impl<'ctx, 'this> Lowerer<'ctx, 'this>
where
    'ctx: 'this,
{
    /// Create a new lowerer. We will lower things into a module.
    pub fn new(context: &'this Context<'ctx>, module: impl AsRef<str>) -> Result<Self, CodegenError> {
        Ok(Self { context, module: Module::new(context, module.as_ref()).map_err(CodegenError::NulError)? })
    }

    pub fn declare_globals<'a>(
        self,
        iter: impl IntoIterator<Item = (impl AsRef<str>, impl AsRef<str>, Vec<&'a ast::TypeInfo>, &'a ast::TypeInfo)>,
    ) -> Result<Self, CodegenError> {
        iter.into_iter()
            .try_fold(self, |this, (module, name, args, ret)| this.declare_global(module, name, args.as_slice(), ret))
    }

    pub fn declare_global(
        mut self,
        m_name: impl AsRef<str>,
        name: impl AsRef<str>,
        args: &[&ast::TypeInfo],
        ret: &ast::TypeInfo,
    ) -> Result<Self, CodegenError> {
        // To avoid code duplication due to the impl thingy
        fn aux(
            c: &Context,
            m: &mut Module,
            m_name: &str,
            name: &str,
            args: &[&ast::TypeInfo],
            ret: &ast::TypeInfo,
        ) -> Result<(), CodegenError> {
            match m.get_function(format!("{m_name}.{name}")).err() {
                Some(LLVMError::Undefined(_)) => m
                    .add_function(
                        VVRTSymbol::new(VVRTSymbolType::Function, m_name, name)
                            .map_err(|err| CodegenError::Error(format!("{err}").into()))?
                            .to_string(),
                        FunctionType::from((TypeLowerer::new(c).lower_slice(args), TypeLowerer::new(c).lower(ret))),
                    )
                    .map_err(CodegenError::NulError)?,
                Some(LLVMError::NulError(err)) => return Err(CodegenError::NulError(err)),
                None => return Err(CodegenError::RedefinedFunction(m_name.into(), name.into())),
            };
            Ok(())
        }

        let Self { context, ref mut module, .. } = self;
        aux(context, module, m_name.as_ref(), name.as_ref(), args, ret).map(|_| self)
    }

    pub fn lower_modules(
        self,
        iter: impl IntoIterator<Item = (impl AsRef<str>, ast::Ast)>,
    ) -> Result<Self, CodegenError> {
        iter.into_iter()
            .try_fold(self, |this, (name, ast)| this.lower_module(name, ast))
    }

    pub fn lower_module(self, name: impl AsRef<str>, ast: ast::Ast) -> Result<Self, CodegenError> {
        todo!("{} -> {ast}", name.as_ref())
    }

    pub fn generate_main(self, main: MainProgram) -> Result<Self, CodegenError> {
        todo!("{main:#?}")
    }

    /// Finished the lowering process and get the module out of the builder.
    pub fn finished(self) -> Module<'ctx> {
        self.module
    }
}
