use codespan_reporting::{
    diagnostic::{Diagnostic, Label},
    files::SimpleFiles,
    term::{Chars, DisplayStyle},
};
use derive_more::Display;
use hashbrown::HashMap;
use std::{mem, str};
use termcolor::ColorChoice;
use vvs_parser::prelude::VVSParserError;

/// Utility to report error to the user.
pub struct ErrorReport<'a> {
    /// The list of parsed files.
    files: SimpleFiles<&'a str, &'a str>,

    /// We cache the ID, so that we can call the [ErrorReport::add_error] and
    /// [ErrorReport::add_warning] multiple times on the same file.
    id_cache: HashMap<&'a str, usize>,

    /// The list of [Error], can be errors or warnings depending on their [ErrorReportType].
    messages: HashMap<usize, Vec<(ErrorReportType, VVSParserError)>>,
}

#[derive(Clone, Copy, PartialEq, Eq, Display)]
enum ErrorReportType {
    #[display("error")]
    Error,

    #[display("warning")]
    Warning,
}

impl<'a> ErrorReport<'a> {
    /// Create a new error report.
    pub fn new() -> Self {
        Default::default()
    }

    fn add<I, T>(&mut self, name: &'a str, src: &'a str, ty: ErrorReportType, iter: I) -> &mut Self
    where
        I: IntoIterator<Item = T>,
        T: Into<VVSParserError>,
    {
        let file_id = self.id_cache.entry(name).or_insert_with(|| self.files.add(name, src));
        self.messages
            .entry(*file_id)
            .or_default()
            .extend(iter.into_iter().map(|error| (ty, error.into())));
        self
    }

    /// Add errors to the report.
    pub fn add_errors<I, T>(&mut self, name: &'a str, source: &'a str, errors: I) -> &mut Self
    where
        I: IntoIterator<Item = T>,
        T: Into<VVSParserError>,
    {
        self.add(name, source, ErrorReportType::Error, errors)
    }

    /// Add warnings to the report.
    pub fn add_warnings<I, T>(&mut self, name: &'a str, source: &'a str, errors: I) -> &mut Self
    where
        I: IntoIterator<Item = T>,
        T: Into<VVSParserError>,
    {
        self.add(name, source, ErrorReportType::Warning, errors)
    }

    /// Report all the stored errors. Returns true if the report contained any error. Return false
    /// if the report was empty or contained only warnings. Note that this operation will empty the
    /// report struct.
    pub fn report(&mut self) -> bool {
        if self.messages.is_empty() {
            return false;
        }

        let Self { files, messages, .. } = mem::take(self);
        let mut output = termcolor::StandardStream::stdout(ColorChoice::Auto);
        let config = codespan_reporting::term::Config {
            display_style: DisplayStyle::Rich,
            chars: Chars::box_drawing(),
            ..Default::default()
        };

        messages.into_iter().fold(false, |acc, (file_id, messages)| {
            messages.iter().fold(false, |acc, (report_type, message)| {
                let diagnostic = match report_type {
                    ErrorReportType::Error => Diagnostic::error(),
                    ErrorReportType::Warning => Diagnostic::warning(),
                }
                .with_message(message.error_message())
                .with_code(message.error_code())
                .with_labels(vec![Label::primary(file_id, message.byte_range())]);

                codespan_reporting::term::emit(&mut output, &config, &files, &diagnostic).is_err()
                    || (*report_type == ErrorReportType::Error || acc)
            }) || acc
        })
    }

    /// Report all stored errors and warnings, but don't say if any error was actually reported.
    /// Like [ErrorReport::report], this operation will empty the report.
    pub fn report_and_ignore(&mut self) {
        self.report();
    }
}

impl ErrorReport<'_> {
    /// Reports a single error
    pub fn error(name: &'_ str, source: &'_ str, error: impl Into<VVSParserError>) {
        ErrorReport::new().add_errors(name, source, [error.into()]).report();
    }

    /// Reports multiple errors
    pub fn errors(name: &'_ str, source: &'_ str, errors: impl IntoIterator<Item = impl Into<VVSParserError>>) {
        ErrorReport::new()
            .add_errors(name, source, errors.into_iter().map(Into::into))
            .report();
    }

    /// Reports a single warning
    pub fn warning(name: &'_ str, source: &'_ str, warning: impl Into<VVSParserError>) {
        ErrorReport::new().add_warnings(name, source, [warning.into()]).report();
    }
}

impl Default for ErrorReport<'_> {
    fn default() -> Self {
        Self { files: SimpleFiles::new(), messages: Default::default(), id_cache: Default::default() }
    }
}
