//! Vivy specific logic

mod error_report;
mod frontend_pipeline;
mod library;
mod main_program;
mod passes;
mod search_path;
mod symbol_table;

use derive_more::Display;
use std::{
    borrow::Cow,
    fs, io,
    path::{Path, PathBuf},
};

pub use crate::{
    error_report::ErrorReport,
    frontend_pipeline::*,
    main_program::*,
    passes::UserFrontendPass,
    search_path::{SearchPath, VIVY_SCRIPT_EXTENSION},
};

/// The representation of a source file.
#[derive(Display, Debug, Clone)]
#[display("{code}")]
pub struct SourceCode {
    path: Option<PathBuf>,
    name: Cow<'static, str>,
    code: Cow<'static, str>,
}

impl SourceCode {
    /// Create a new source file from a file.
    pub fn from_path(path: impl AsRef<Path>) -> io::Result<Self> {
        let path = path.as_ref();
        Ok(Self {
            name: Cow::Owned(
                path.file_name()
                    .ok_or_else(|| io::Error::new(io::ErrorKind::Other, "path doesn't have a file name"))?
                    .to_string_lossy()
                    .to_string(),
            ),
            path: Some(path.to_path_buf()),
            code: Cow::Owned(fs::read_to_string(path)?),
        })
    }

    /// Create a new source file from memory.
    pub fn from_memory(code: Cow<'static, str>) -> Self {
        Self { name: Cow::Borrowed("--"), code, path: None }
    }

    /// Get the name of the source file.
    pub fn name(&self) -> &str {
        &self.name
    }

    /// Get the name of the module. If the source code doesn't have a valid name (e.g. it doesn't
    /// end in ".vvs"), returns [None].
    pub fn module(&self) -> Option<&str> {
        self.name().strip_suffix(VIVY_SCRIPT_EXTENSION)?.strip_suffix('.')
    }

    /// Gets the parent folder of the script if it exists.
    pub fn folder(&self) -> Option<&Path> {
        self.path.as_deref()
    }
}
