//! Main program and utilities to be able to complete the work.

use derive_more::{Display, IntoIterator, IsVariant};
use std::fmt;
use vvs_parser::prelude::ast::Expression;
use vvs_shortstring::ShortString;

/// The main program to execute in the parsed script.
#[derive(Debug)]
pub struct MainProgram {
    /// The name of the program, i.e. the main module.
    pub name: ShortString,

    /// The initial variable name of the read ASS file.
    pub initial: ShortString,

    /// Steps to do, in order, to make the program progress.
    pub steps: Vec<MainProgramStep>,

    /// The variables to write at the end of the program.
    pub returns: MainProgramReturns,
}

impl Default for MainProgram {
    fn default() -> Self {
        Self {
            name: ShortString::new_static("stdin"),
            initial: ShortString::new_static("init"),
            steps: Default::default(),
            returns: Default::default(),
        }
    }
}

impl MainProgram {
    /// Set the name of the progeam.
    pub fn with_name(self, name: ShortString) -> Self {
        Self { name, ..self }
    }

    /// Set the initial variable name.
    pub fn with_initial(self, initial: ShortString) -> Self {
        Self { initial, ..self }
    }

    /// Set the steps for the main program.
    pub fn with_steps(self, steps: Vec<MainProgramStep>) -> Self {
        Self { steps, ..self }
    }

    /// Set the returned variables
    pub fn with_returns(self, returns: impl IntoIterator<Item = ShortString>) -> Self {
        Self { returns: returns.into_iter().collect(), ..self }
    }
}

impl fmt::Display for MainProgram {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        let Self { name, initial, steps, returns } = self;
        writeln!(f, "program {name} ({initial}) {{")?;
        steps.iter().try_for_each(|step| writeln!(f, "{step}"))?;
        writeln!(f, "\t{returns}")?;
        write!(f, "}}")
    }
}

/// The returned variables from the main program.
#[derive(Debug, Default, IntoIterator)]
pub struct MainProgramReturns {
    #[into_iterator(owned)]
    variables: Vec<ShortString>,
}

impl FromIterator<ShortString> for MainProgramReturns {
    fn from_iter<T: IntoIterator<Item = ShortString>>(iter: T) -> Self {
        Self { variables: iter.into_iter().collect() }
    }
}

impl fmt::Display for MainProgramReturns {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "return {{")?;
        match &self.variables[..] {
            [] => write!(f, " }}"),
            [item] => write!(f, " {item} }}"),
            [head, tail @ ..] => {
                write!(f, " {head}")?;
                tail.iter().try_for_each(|item| write!(f, ", {item}"))?;
                write!(f, " }}")
            }
        }
    }
}

/// A step to do for the main program.
#[derive(Debug)]
pub struct MainProgramStep {
    /// The name of the assignated variable.
    destination: ShortString,

    /// The name of the job with its source module, in this order: `(module, job)`.
    job: (ShortString, ShortString),

    /// The arguments to pass to the job, the variables to read (and possibly rearange), and the
    /// parameters to set.
    call_list: Vec<MainProgramCallListItem>,
}

impl MainProgramStep {
    /// Create a new program step.
    pub fn new(
        destination: ShortString,
        job: (ShortString, ShortString),
        call_list: impl IntoIterator<Item = MainProgramCallListItem>,
    ) -> Self {
        Self { destination, job, call_list: call_list.into_iter().collect() }
    }

    /// Get the name of the variable to write the result of the job into.
    pub fn destination(&self) -> &ShortString {
        &self.destination
    }

    /// Get the name of the source module where to find the job we are trying to call.
    pub fn module(&self) -> &ShortString {
        &self.job.0
    }

    /// Get the name of the job we are trying to call.
    pub fn job(&self) -> &ShortString {
        &self.job.1
    }

    /// Get the arguments of the job.
    pub fn call_list_items(&self) -> impl Iterator<Item = &MainProgramCallListItem> {
        self.call_list.iter()
    }
}

impl fmt::Display for MainProgramStep {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        let Self { destination, job: (module, job), .. } = self;
        write!(f, "{destination} = {module}:{job}")?;
        match &self.call_list[..] {
            [] => write!(f, " {{ }}"),
            [item] => write!(f, " {{ {item} }}"),
            [head, tail @ ..] => {
                write!(f, " {{ {head}")?;
                tail.iter().try_for_each(|item| write!(f, ", {item}"))?;
                write!(f, " }}")
            }
        }
    }
}

/// An item to pass to a job when we call it.
#[derive(Debug, IsVariant, Display)]
pub enum MainProgramCallListItem {
    /// We add a variable to process.
    #[display("{_0}")]
    Variable(ShortString),

    /// We set a parameter for the job.
    #[display("{_0} = {_1}")]
    Parameter(ShortString, Expression),
}

impl MainProgramCallListItem {
    /// Treat the call list item as a variable if possible.
    pub fn as_variable(&self) -> Option<&ShortString> {
        match self {
            MainProgramCallListItem::Variable(variable) => Some(variable),
            _ => None,
        }
    }

    /// Treat the call list item as a parameter setting if possible.
    pub fn as_parameter(&self) -> Option<(&ShortString, &Expression)> {
        match self {
            MainProgramCallListItem::Parameter(argument, value) => Some((argument, value)),
            _ => None,
        }
    }
}
