use crate::prelude::*;
use llvm_sys::{core::*, prelude::*};

crate::bindings::declare! { const LLVMBasicBlockRef as BB<'a>    }
crate::bindings::declare! { mut   LLVMBasicBlockRef as BBMut<'a> {
    const_ref: BB<'a>,
} }

/// Share code between the [BB::terminator] and [BB::terminator_mut]
macro_rules! get_terminator {
    ($self:ident as $value:ident) => {{
        use llvm_sys::LLVMOpcode::*;
        match unsafe { LLVMGetInstructionOpcode(LLVMGetLastInstruction($self.as_ptr())) } {
            LLVMRet | LLVMBr | LLVMSwitch | LLVMIndirectBr | LLVMCallBr => unsafe {
                Some($value::from_ptr(LLVMGetBasicBlockTerminator($self.as_ptr())))
            },

            LLVMCleanupRet | LLVMCatchRet | LLVMCatchPad | LLVMCleanupPad | LLVMCatchSwitch => unsafe {
                Some($value::from_ptr(LLVMGetBasicBlockTerminator($self.as_ptr())))
            },

            LLVMUnreachable => None,

            _ => None,
        }
    }};
}

impl<'a> BBMut<'a> {
    /// Get a const reference to the basic block.
    pub fn into_ref(self) -> BB<'a> {
        unsafe { BB::from_ptr(self.as_ptr()) }
    }

    /// Get the terminator in a mutable way of the basic block if it exists.
    pub fn terminator_mut(&mut self) -> Option<ValueMut> {
        get_terminator!(self as ValueMut)
    }

    /// Iterate over the instructions of a basic block in a mutable way.
    pub fn iter_mut(&'a mut self) -> BBIterMut<'a, &'a mut BBMut<'a>> {
        match self.as_ref().is_empty() {
            true => BBIterMut { curr: None, last: None, bb: self },
            false => todo!(),
        }
    }

    /// Create a new [BBMut] from the pointer.
    pub(crate) unsafe fn new(ptr: LLVMBasicBlockRef) -> Self {
        unsafe { Self::from_ptr(ptr, BB::from_ptr(ptr)) }
    }
}

impl<'a> BB<'a> {
    /// Get the insturction count of this basic block.
    pub fn len(&self) -> usize {
        (unsafe { LLVMGetLastInstruction(self.as_ptr()).offset_from(LLVMGetFirstInstruction(self.as_ptr())) }) as usize
    }

    /// Tells wether there are any instructions in this basic block.
    pub fn is_empty(&self) -> bool {
        self.len() == 0
    }

    /// Get the terminator of the basic block if it exists.
    pub fn terminator(&self) -> Option<Value> {
        get_terminator!(self as Value)
    }

    /// Iterate over the instructions of a basic block.
    pub fn iter(&'a self) -> BBIter<'a, &'a BB<'a>> {
        match self.is_empty() {
            true => BBIter { curr: None, last: None, bb: self },
            false => todo!(),
        }
    }

    /// Create a new [BB] from the pointer.
    pub(crate) unsafe fn new(ptr: LLVMBasicBlockRef) -> Self {
        unsafe { Self::from_ptr(ptr) }
    }
}

impl<'a> AsRef<BB<'a>> for BBMut<'a> {
    fn as_ref(&self) -> &BB<'a> {
        &self.const_ref
    }
}

impl<'a> AsRef<BB<'a>> for &'a BB<'a> {
    fn as_ref(&self) -> &BB<'a> {
        self
    }
}

impl<'a> AsRef<BB<'a>> for &'a mut BB<'a> {
    fn as_ref(&self) -> &BB<'a> {
        self
    }
}

impl<'a> AsMut<BBMut<'a>> for &'a mut BBMut<'a> {
    fn as_mut(&mut self) -> &mut BBMut<'a> {
        self
    }
}

impl Eq for BB<'_> {}
impl PartialEq for BB<'_> {
    fn eq(&self, other: &Self) -> bool {
        self.inner == other.inner
    }
}

impl Eq for BBMut<'_> {}
impl PartialEq for BBMut<'_> {
    fn eq(&self, other: &Self) -> bool {
        self.inner == other.inner
    }
}

mod sealed {
    pub trait BBRef<'a>: AsRef<super::BB<'a>> {}
    pub trait BBMutRef<'a>: AsMut<super::BBMut<'a>> + AsRef<super::BB<'a>> {}
}

impl<'a> sealed::BBRef<'a> for &'a BB<'a> {}
impl<'a> sealed::BBRef<'a> for &'a mut BB<'a> {}
impl<'a> sealed::BBMutRef<'a> for &'a mut BBMut<'a> {}

/// Iterate over the instructions in a basic block.
pub struct BBIter<'a, B: sealed::BBRef<'a>> {
    curr: Option<Value<'a>>,
    last: Option<Value<'a>>,
    bb: B,
}

/// Iterate over the instructions in a basic block.
pub struct BBIterMut<'a, B: sealed::BBMutRef<'a>> {
    curr: Option<ValueMut<'a>>,
    last: Option<ValueMut<'a>>,
    bb: B,
}

macro_rules! next {
    ($self:ident as $value:ident) => {
        match ($self.curr.take(), $self.last.as_ref()) {
            (Some(_), None) | (None, Some(_)) => unreachable!(),
            (None, None) => None,
            (Some(curr), Some(last)) if curr == *last => {
                $self.last = None;
                Some(curr)
            }
            (Some(curr), Some(_)) => unsafe {
                $self.curr = Some($value::from_ptr(LLVMGetNextInstruction(curr.as_ptr())));
                Some(curr)
            },
        }
    };
}

impl<'a, B: sealed::BBRef<'a>> Iterator for BBIter<'a, B> {
    type Item = Value<'a>;

    fn next(&mut self) -> Option<Self::Item> {
        next!(self as Value)
    }

    fn last(self) -> Option<Self::Item> {
        self.last
    }

    fn count(self) -> usize {
        self.size_hint().0
    }

    fn size_hint(&self) -> (usize, Option<usize>) {
        let size = self.bb.as_ref().len();
        (size, Some(size))
    }
}

impl<'a, B: sealed::BBMutRef<'a>> Iterator for BBIterMut<'a, B> {
    type Item = ValueMut<'a>;

    fn next(&mut self) -> Option<Self::Item> {
        next!(self as ValueMut)
    }

    fn last(self) -> Option<Self::Item> {
        self.last
    }

    fn count(self) -> usize {
        self.size_hint().0
    }

    fn size_hint(&self) -> (usize, Option<usize>) {
        let size = self.bb.as_ref().len();
        (size, Some(size))
    }
}
