use crate::{bindings::cstr, prelude::*};
use llvm_sys::{core::*, prelude::*, LLVMLinkage, LLVMVisibility};

crate::bindings::declare! { const LLVMValueRef as FunctionDeclaration<'a> }
crate::bindings::declare! { mut   LLVMValueRef as FunctionBuilder<'a> }
crate::bindings::declare! { mut   LLVMValueRef as Function<'a> {
    entry_point: BB<'a>,
    basic_block_count: u64,
} }

impl<'a> FunctionDeclaration<'a> {
    /// Get the type of this function declaration.
    pub fn ty(&self) -> FunctionType {
        unsafe { FunctionType::from_ptr(LLVMTypeOf(self.as_ptr())) }
    }

    /// Get the LLVM context of the function.
    pub fn context(&'a self) -> Context<'a> {
        unsafe { Context::from_ptr(LLVMGetTypeContext(LLVMTypeOf(self.inner)), true) }
    }
}

impl<'a> FunctionBuilder<'a> {
    /// Get the associated declaration.
    pub fn declaration(&self) -> FunctionDeclaration {
        unsafe { FunctionDeclaration::from_ptr(self.as_ptr()) }
    }

    /// Set the [LLVMLinkage] and [LLVMVisibility] accordingly.
    pub fn into_public(self, public: bool) -> Self {
        match public {
            true => unsafe {
                LLVMSetLinkage(self.as_ptr(), LLVMLinkage::LLVMCommonLinkage);
                LLVMSetVisibility(self.as_ptr(), LLVMVisibility::LLVMProtectedVisibility);
            },
            false => unsafe {
                LLVMSetLinkage(self.as_ptr(), LLVMLinkage::LLVMLinkerPrivateLinkage);
                LLVMSetVisibility(self.as_ptr(), LLVMVisibility::LLVMHiddenVisibility);
            },
        }
        self
    }

    /// Transform the function declaration into a proper function, with a function body, or at
    /// lease an entry point.
    pub fn into_function(self) -> Function<'a> {
        unsafe {
            Function::from_ptr(
                self.as_ptr(),
                BB::from_ptr(LLVMAppendBasicBlockInContext(
                    self.declaration().context().as_ptr(),
                    self.as_ptr(),
                    crate::bindings::cstr(b"entrypoint\0").as_ptr(),
                )),
                0,
            )
        }
    }
}

impl<'a> Function<'a> {
    /// Get the associated declaration.
    pub fn declaration(&self) -> FunctionDeclaration {
        unsafe { FunctionDeclaration::from_ptr(self.as_ptr()) }
    }

    /// Get the entry point of the function.
    pub fn entry_point(&self) -> &BB {
        &self.entry_point
    }

    /// Get the entry point of the function in a mutable way.
    pub fn entry_point_mut(&'a mut self) -> &'a mut BB<'a> {
        &mut self.entry_point
    }

    /// Add a new basic block to the function and returns it.
    pub fn new_basic_block(&'a mut self) -> BB<'a> {
        self.basic_block_count += 1;
        unsafe {
            BB::from_ptr(LLVMAppendBasicBlockInContext(
                self.declaration().context().as_ptr(),
                self.inner,
                cstr(format!("{}", self.basic_block_count).as_bytes()).as_ptr(),
            ))
        }
    }

    /// Iterate over all the basic blocks of the function.
    pub fn iter(&'a self) -> FunctionIter<'a, &'a Function<'a>> {
        unsafe {
            match LLVMCountBasicBlocks(self.as_ptr()) {
                0 => FunctionIter { curr: None, last: None, func: self },
                _ => FunctionIter {
                    curr: Some(BB::new(LLVMGetFirstBasicBlock(self.inner))),
                    last: Some(BB::new(LLVMGetLastBasicBlock(self.inner))),
                    func: self,
                },
            }
        }
    }

    /// Iterate over all the basic blocks of the function in a mutable way.
    pub fn iter_mut(&'a mut self) -> FunctionIterMut<'a, &'a mut Function<'a>> {
        unsafe {
            match LLVMCountBasicBlocks(self.as_ptr()) {
                0 => FunctionIterMut { curr: None, last: None, func: self },
                _ => FunctionIterMut {
                    curr: Some(BBMut::new(LLVMGetFirstBasicBlock(self.inner))),
                    last: Some(BBMut::new(LLVMGetLastBasicBlock(self.inner))),
                    func: self,
                },
            }
        }
    }
}

mod sealed {
    pub trait FuncRef<'a>: AsRef<super::Function<'a>> {}
}

impl<'a> sealed::FuncRef<'a> for &'a Function<'a> {}
impl<'a> sealed::FuncRef<'a> for &'a mut Function<'a> {}

impl<'a> AsRef<Function<'a>> for &'a Function<'a> {
    fn as_ref(&self) -> &Function<'a> {
        self
    }
}

impl<'a> AsRef<Function<'a>> for &'a mut Function<'a> {
    fn as_ref(&self) -> &Function<'a> {
        self
    }
}

/// Iterate over the basic blocks in a function.
pub struct FunctionIter<'a, F: sealed::FuncRef<'a>> {
    curr: Option<BB<'a>>,
    last: Option<BB<'a>>,
    func: F,
}

/// Iterate over the basic blocks in a function.
pub struct FunctionIterMut<'a, F: sealed::FuncRef<'a>> {
    curr: Option<BBMut<'a>>,
    last: Option<BBMut<'a>>,
    func: F,
}

macro_rules! next {
    ($self:ident as $bb:ident) => {
        match ($self.curr.take(), $self.last.as_ref()) {
            (Some(_), None) | (None, Some(_)) => unreachable!(),
            (None, None) => None,
            (Some(curr), Some(last)) if curr == *last => {
                $self.last = None;
                Some(curr)
            }
            (Some(curr), Some(_)) => unsafe {
                $self.curr = Some($bb::new(LLVMGetNextBasicBlock(curr.as_ptr())));
                Some(curr)
            },
        }
    };
}

impl<'a, F: sealed::FuncRef<'a>> Iterator for FunctionIter<'a, F> {
    type Item = BB<'a>;

    fn next(&mut self) -> Option<Self::Item> {
        next!(self as BB)
    }

    fn last(self) -> Option<Self::Item> {
        self.last
    }

    fn count(self) -> usize {
        (unsafe { LLVMCountBasicBlocks(self.func.as_ref().as_ptr()) }) as usize
    }

    fn size_hint(&self) -> (usize, Option<usize>) {
        let count = (unsafe { LLVMCountBasicBlocks(self.func.as_ref().as_ptr()) }) as usize;
        (count, Some(count))
    }
}

impl<'a, F: sealed::FuncRef<'a>> Iterator for FunctionIterMut<'a, F> {
    type Item = BBMut<'a>;

    fn next(&mut self) -> Option<Self::Item> {
        next!(self as BBMut)
    }

    fn last(self) -> Option<Self::Item> {
        self.last
    }

    fn count(self) -> usize {
        (unsafe { LLVMCountBasicBlocks(self.func.as_ref().as_ptr()) }) as usize
    }

    fn size_hint(&self) -> (usize, Option<usize>) {
        let count = (unsafe { LLVMCountBasicBlocks(self.func.as_ref().as_ptr()) }) as usize;
        (count, Some(count))
    }
}
