use crate::{
    tokenizer::{Lexer, LexerResult, Symbol, TokenKind, TokenReference},
    traits::MaybeLazyString,
};
use std::borrow::Cow;

mod ast_result;
mod option_table_result;
mod result;

pub use crate::ast::parsers::structs::{ast_result::*, option_table_result::*, result::*};

pub struct ParserState {
    errors: Vec<crate::Error>,
    lexer: Lexer,
}

impl ParserState {
    pub fn new(lexer: Lexer) -> Self {
        Self { errors: Vec::new(), lexer }
    }

    pub fn current(&self) -> Result<&TokenReference, ()> {
        match self.lexer.current() {
            Some(LexerResult::Ok(token) | LexerResult::Recovered(token, _)) => Ok(token),
            Some(LexerResult::Fatal(_)) => Err(()),
            None => unreachable!("current() called past EOF"),
        }
    }

    pub fn peek(&self) -> Result<&TokenReference, ()> {
        match self.lexer.peek() {
            Some(LexerResult::Ok(token) | LexerResult::Recovered(token, _)) => Ok(token),
            Some(LexerResult::Fatal(_)) => Err(()),
            None => unreachable!("peek() called past EOF"),
        }
    }

    pub fn consume(&mut self) -> ParserResult<TokenReference> {
        match self.lexer.consume() {
            Some(LexerResult::Ok(token)) => ParserResult::Value(token),
            Some(LexerResult::Recovered(token, errors)) => {
                self.errors.extend(errors.into_iter().map(crate::Error::TokenizerError));
                ParserResult::Value(token)
            }
            Some(LexerResult::Fatal(errors)) => {
                self.errors.extend(errors.into_iter().map(crate::Error::TokenizerError));
                ParserResult::LexerMoved
            }
            None => ParserResult::NotFound,
        }
    }

    pub fn consume_ifs(&mut self, symbols: &[Symbol]) -> Option<TokenReference> {
        symbols.iter().copied().find_map(|symbol| self.consume_if(symbol))
    }

    pub fn consume_if(&mut self, symbol: Symbol) -> Option<TokenReference> {
        self.current().ok()?.is_symbol(symbol).then(|| self.consume().unwrap())
    }

    pub fn consume_kind(&mut self, kind: TokenKind) -> Option<TokenReference> {
        if self.current().ok()?.is_kind(kind) {
            Some(self.consume().unwrap())
        } else {
            self.token_error(self.current().unwrap().clone(), format!("expected a token of type {kind:?}"));
            None
        }
    }

    pub fn require(&mut self, symbol: Symbol, error: &'static str) -> Option<TokenReference> {
        if self.current().ok()?.is_symbol(symbol) {
            Some(self.consume().unwrap())
        } else {
            self.token_error(self.current().ok()?.clone(), error);
            None
        }
    }

    pub fn require_with_reference_token(
        &mut self,
        symbol: Symbol,
        error: &'static str,
        reference_token: &TokenReference,
    ) -> Option<TokenReference> {
        if self.current().ok()?.is_symbol(symbol) {
            Some(self.consume().unwrap())
        } else {
            self.token_error(reference_token.clone(), error);
            None
        }
    }

    pub fn require_with_reference_range(
        &mut self,
        symbol: Symbol,
        error: impl MaybeLazyString,
        start_token: &TokenReference,
        end_token: &TokenReference,
    ) -> Option<TokenReference> {
        let token = self.current().ok()?;
        if token.is_symbol(symbol) {
            Some(self.consume().unwrap())
        } else {
            self.token_error_ranged(token.clone(), error.to_str(), start_token, end_token);
            None
        }
    }

    pub fn require_with_reference_range_callback(
        &mut self,
        symbol: Symbol,
        error: impl MaybeLazyString,
        tokens: impl FnOnce() -> (TokenReference, TokenReference),
    ) -> Option<TokenReference> {
        let token = self.current().ok()?;
        if token.is_symbol(symbol) {
            Some(self.consume().unwrap())
        } else {
            let (start_token, end_token) = tokens();
            self.token_error_ranged(token.clone(), error.to_str(), &start_token, &end_token);
            None
        }
    }

    pub fn token_error(&mut self, token_reference: TokenReference, error: impl Into<Cow<'static, str>>) {
        self.errors.push(crate::Error::AstError(crate::ast::AstError {
            token: token_reference.token,
            additional: error.into(),
            range: None,
        }));
    }

    // This takes start_token and end_token as owned references because otherwise, we tend to stack an immutable over mutable borrow.
    pub fn token_error_ranged(
        &mut self,
        token_reference: TokenReference,
        error: impl Into<Cow<'static, str>>,
        start_token: &TokenReference,
        end_token: &TokenReference,
    ) {
        self.errors.push(crate::Error::AstError(crate::ast::AstError {
            token: token_reference.token,
            additional: error.into(),
            range: Some((start_token.start_position(), end_token.end_position())),
        }));
    }
}
