//! Special errors for the Vivy Script package, so we don't expose anything with anyhow.

use crate::{
    ast::AstError,
    tokenizer::{Position, TokenizerError},
};
use derive_more::{Display, From};
use serde::{Deserialize, Serialize};
use std::{borrow::Cow, ops, rc::Rc};

/// An error type that consists of both [`AstError`](ast::AstError) and [`TokenizerError`](tokenizer::TokenizerError)
/// Used by [`parse`]
#[derive(Clone, Debug, PartialEq, Eq, Deserialize, Serialize, From, Display)]
pub enum Error {
    /// Triggered if there's an issue creating an AST, but tokenizing must have succeeded
    #[display("error occurred while creating ast: {_0}")]
    AstError(#[from] AstError),

    /// Triggered if there's an issue when tokenizing, and an AST can't be made
    #[display("error occurred while tokenizing: {_0}")]
    TokenizerError(#[from] TokenizerError),
}

impl From<Box<AstError>> for Error {
    fn from(value: Box<AstError>) -> Self {
        Error::from(*value)
    }
}

impl From<Rc<AstError>> for Error {
    fn from(value: Rc<AstError>) -> Self {
        Error::from(Rc::unwrap_or_clone(value))
    }
}

impl Error {
    /// Returns a human readable error message
    pub fn error_message(&self) -> Cow<'static, str> {
        match self {
            Error::AstError(error) => error.error_message(),
            Error::TokenizerError(error) => error.to_string().into(),
        }
    }

    /// Returns the range of the error
    pub fn range(&self) -> (Position, Position) {
        match self {
            Error::AstError(error) => error.range(),
            Error::TokenizerError(error) => error.range(),
        }
    }

    /// Returns the range of the error in bytes
    pub fn byte_range(&self) -> ops::Range<usize> {
        let (start, end) = match self {
            Error::AstError(error) => error.range(),
            Error::TokenizerError(error) => error.range(),
        };
        start.bytes()..end.bytes()
    }

    /// Get the error code.
    pub fn error_code(&self) -> &'static str {
        match self {
            Error::AstError(_) => "ast",
            Error::TokenizerError(_) => "tokenizer",
        }
    }
}

impl std::error::Error for Error {}
