//! Build the tokens. Don't need for outter world to know the lexer.

use crate::{
    tokenizer::{
        Position, StringLiteralQuoteType, Symbol, Token, TokenReference, TokenType, TokenizerError, TokenizerErrorType,
    },
    ShortString,
};
use serde::{Deserialize, Serialize};

/// A lexer, which will produce a stream of tokens from a source string.
/// If you just want to create an [`Ast`](crate::ast::Ast) from a string, you want to use
/// [`parse`](crate::parse) instead.
pub struct Lexer {
    pub(crate) source: LexerSource,
    sent_eof: bool,

    next_token: Option<LexerResult<TokenReference>>,
    peek_token: Option<LexerResult<TokenReference>>,
}

impl Lexer {
    /// Creates a new Lexer from the given source string.
    pub fn new(source: &str) -> Self {
        let mut lexer = Self::new_lazy(source);
        lexer.next_token = lexer.process_first_with_trivia();
        lexer.peek_token = lexer.process_next_with_trivia();
        lexer
    }

    /// Creates a new Lexer from the given source string, but does not process
    /// the first token.
    pub fn new_lazy(source: &str) -> Self {
        Self { source: LexerSource::new(source), sent_eof: false, next_token: None, peek_token: None }
    }

    /// Returns the current token.
    pub fn current(&self) -> Option<&LexerResult<TokenReference>> {
        self.next_token.as_ref()
    }

    /// Returns the next token.
    pub fn peek(&self) -> Option<&LexerResult<TokenReference>> {
        self.peek_token.as_ref()
    }

    /// Consumes the current token and returns the next token.
    pub fn consume(&mut self) -> Option<LexerResult<TokenReference>> {
        let next = self.next_token.take()?;
        self.next_token = self.peek_token.take();
        self.peek_token = self.process_next_with_trivia();
        Some(next)
    }

    /// Returns a vector of all tokens left in the source string.
    #[allow(dead_code)]
    pub fn collect(self) -> LexerResult<Vec<Token>> {
        let mut tokens = Vec::new();
        let mut lexer = self;
        let mut errors = Vec::new();

        while let Some(token_reference) = lexer.consume() {
            let mut token_reference = match token_reference {
                LexerResult::Ok(token_reference) => token_reference,
                LexerResult::Recovered(token_reference, mut new_errors) => {
                    errors.append(&mut new_errors);
                    token_reference
                }
                LexerResult::Fatal(mut new_errors) => {
                    errors.append(&mut new_errors);
                    continue;
                }
            };
            tokens.append(&mut token_reference.leading_trivia);
            tokens.push(token_reference.token);
            tokens.append(&mut token_reference.trailing_trivia);
        }

        LexerResult::new(tokens, errors)
    }

    fn create(&self, start_position: Position, token_type: TokenType) -> Option<LexerResult<Token>> {
        Some(LexerResult::Ok(Token { token_type, start_position, end_position: self.source.position() }))
    }

    fn create_recovered(
        &self,
        start_position: Position,
        token_type: TokenType,
        errors: Vec<TokenizerError>,
    ) -> Option<LexerResult<Token>> {
        Some(LexerResult::new(Token { token_type, start_position, end_position: self.source.position() }, errors))
    }

    fn process_next_with_trivia(&mut self) -> Option<LexerResult<TokenReference>> {
        let mut leading_trivia = Vec::new();
        let mut errors: Option<Vec<TokenizerError>> = None;

        let nontrivial_token = loop {
            match self.process_next()? {
                LexerResult::Ok(token) if token.token_type().is_trivia() => leading_trivia.push(token),
                LexerResult::Ok(token) => break token,
                LexerResult::Fatal(error) => return Some(LexerResult::Fatal(error)),
                LexerResult::Recovered(token, mut new_errors) => {
                    match errors.as_mut() {
                        Some(errors) => errors.append(&mut new_errors),
                        None => errors = Some(new_errors),
                    }
                    if token.token_type().is_trivia() {
                        leading_trivia.push(token);
                    } else {
                        break token;
                    }
                }
            }
        };

        let trailing_trivia = self.collect_trailing_trivia();
        let token = TokenReference { token: nontrivial_token, leading_trivia, trailing_trivia };
        match errors {
            Some(errors) => Some(LexerResult::Recovered(token, errors)),
            None => Some(LexerResult::Ok(token)),
        }
    }

    fn process_first_with_trivia(&mut self) -> Option<LexerResult<TokenReference>> {
        if self.source.current() == Some('#') && self.source.peek() == Some('!') {
            let start_position = self.source.position();
            let mut line = "#!".to_string();

            self.source.next();
            self.source.next();

            while let Some(next) = self.source.current() {
                if next == '\n' {
                    break;
                }
                self.source.next();
                line.push(next);
            }

            let end_position = self.source.position();
            let shebang = Token { token_type: TokenType::Shebang { line: line.into() }, start_position, end_position };

            // rewrite todo: handle LexerResult better here
            if let Some(LexerResult::Ok(mut token_reference)) = self.process_next_with_trivia() {
                token_reference.leading_trivia.insert(0, shebang);
                return Some(LexerResult::Ok(token_reference));
            }
        }

        self.process_next_with_trivia()
    }

    fn collect_trailing_trivia(&mut self) -> Vec<Token> {
        let mut trailing_trivia = Vec::new();

        loop {
            let sent_eof = self.sent_eof;
            let start_position = self.source.lexer_position;

            match self.process_next() {
                Some(LexerResult::Ok(token)) if token.token_type().is_trivia() => {
                    // Take all trivia up to and including the newline character. If we see a newline character
                    // we should break once we have taken it in.
                    let should_break = if let TokenType::Whitespace { ref characters } = token.token_type() {
                        // Use contains in order to tolerate \r\n line endings and mixed whitespace tokens
                        characters.contains('\n')
                    } else {
                        false
                    };
                    trailing_trivia.push(token);
                    if should_break {
                        break;
                    }
                }

                _ => {
                    self.source.lexer_position = start_position;
                    self.sent_eof = sent_eof;
                    break;
                }
            }
        }

        trailing_trivia
    }

    /// Processes and returns the next token in the source string, ignoring trivia.
    pub fn process_next(&mut self) -> Option<LexerResult<Token>> {
        let start_position = self.source.position();
        let Some(next) = self.source.next() else {
            if self.sent_eof {
                return None;
            } else {
                self.sent_eof = true;
                return self.create(start_position, TokenType::Eof);
            }
        };

        macro_rules! consume {
            ($first: literal, $second: literal) => {
                self.source.consume_two($first, $second)
            };
            ($first: literal) => {
                self.source.consume($first)
            };
        }
        macro_rules! create {
            ($symbol: ident) => {
                self.create(start_position, TokenType::Symbol { symbol: Symbol::$symbol })
            };
        }

        match next {
            // Identifiers
            initial if is_identifier_start(initial) => {
                let mut identifier = String::new();
                identifier.push(initial);

                while let Some(next) = self.source.current() {
                    if is_identifier_start(next) || next.is_ascii_digit() {
                        identifier.push(self.source.next().expect("peeked, but no next"));
                    } else {
                        break;
                    }
                }

                let token_type = if let Some(symbol) = Symbol::from_str(&identifier) {
                    TokenType::Symbol { symbol }
                } else {
                    TokenType::Identifier { identifier: ShortString::from(identifier) }
                };
                self.create(start_position, token_type)
            }

            // Spaces
            initial @ (' ' | '\t' | '\r') => {
                let mut whitespace = String::new();
                whitespace.push(initial);

                // Handle end_position appropriately: for a newline, we increment the bytes, but
                // do not increment line/char
                let mut end_position = Position {
                    bytes: start_position.bytes() + initial.len_utf8(),
                    line: start_position.line,
                    character: start_position.character + 1,
                };

                while let Some(next) = self.source.current() {
                    if next == ' ' || next == '\t' {
                        end_position.bytes += next.len_utf8();
                        end_position.character += 1;
                        whitespace.push(self.source.next().expect("peeked, but no next"));
                        continue;
                    }

                    if next == '\n' {
                        end_position.bytes += next.len_utf8();
                        whitespace.push(self.source.next().expect("peeked, but no next"));
                    } else if next == '\r' && self.source.peek() == Some('\n') {
                        let carriage_return = self.source.next().expect("peeked, but no next");
                        let new_line = self.source.next().expect("peeked, but no next");
                        end_position.bytes += carriage_return.len_utf8() + new_line.len_utf8();
                        end_position.character += 1;
                        whitespace.push(carriage_return);
                        whitespace.push(new_line);
                    }
                    break;
                }

                Some(LexerResult::Ok(Token {
                    token_type: TokenType::Whitespace { characters: ShortString::from(whitespace.as_str()) },
                    start_position,
                    end_position,
                }))
            }

            // Integers
            '0' if matches!(self.source.current(), Some('x' | 'X')) => {
                let hex_character = self.source.next().unwrap();
                self.read_hex_number(hex_character, start_position)
            }
            '0' if matches!(self.source.current(), Some('b' | 'B')) => {
                let binary_character = self.source.next().unwrap();
                self.read_binary_number(binary_character, start_position)
            }
            initial @ '0' => self.read_number(start_position, initial.to_string()),
            initial @ ('1'..='9') => self.read_number(start_position, initial.to_string()),

            // Strings
            quote @ ('"' | '\'') => {
                let (string, recovered) = self.read_string(quote);
                let errors = recovered.then(|| TokenizerError {
                    error: TokenizerErrorType::UnclosedString,
                    range: (start_position, self.source.position()),
                });
                self.create_recovered(start_position, string, errors.into_iter().collect())
            }

            // Complicated things
            '[' if matches!(self.source.current(), Some('[' | '=')) => {
                let start_lexer_position = self.source.lexer_position;
                match self.read_multi_line_body() {
                    MultiLineBodyResult::Ok { blocks, body } => self.create(
                        start_position,
                        TokenType::StringLiteral {
                            literal: body.into(),
                            multi_line_depth: blocks,
                            quote_type: StringLiteralQuoteType::Brackets,
                        },
                    ),
                    MultiLineBodyResult::Unclosed { blocks, body } => self.create_recovered(
                        start_position,
                        TokenType::StringLiteral {
                            literal: body.into(),
                            multi_line_depth: blocks,
                            quote_type: StringLiteralQuoteType::Brackets,
                        },
                        vec![TokenizerError {
                            error: TokenizerErrorType::UnclosedString,
                            range: (start_position, self.source.position()),
                        }],
                    ),
                    MultiLineBodyResult::NotMultiLine { .. } => {
                        // Reset back, parse the one `[`, and let the rest be parsed again
                        self.source.lexer_position = start_lexer_position;
                        create!(LeftBracket)
                    }
                }
            }

            // Consume three
            '/' if consume!('/', '=') => create!(DoubleSlashEqual),
            '.' if consume!('.', '=') => create!(TwoDotsEqual),

            // Consume two
            '/' if consume!('/') => create!(DoubleSlash),
            '/' if consume!('=') => create!(SlashEqual),
            ':' if consume!(':') => create!(TwoColons),
            '+' if consume!('=') => create!(PlusEqual),
            '*' if consume!('=') => create!(StarEqual),
            '%' if consume!('=') => create!(PercentEqual),
            '^' if consume!('=') => create!(CaretEqual),
            '=' if consume!('=') => create!(TwoEqual),
            '~' if consume!('>') => create!(DoubleGreaterThan),
            '~' if consume!('=') => create!(TildeEqual),
            '<' if consume!('~') => create!(DoubleLesserThan),
            '<' if consume!('=') => create!(LessThanEqual),
            '>' if consume!('=') => create!(GreaterThanEqual),
            '.' if consume!('.') => create!(TwoDots),
            '-' if consume!('=') => create!(MinusEqual),
            '-' if consume!('>') => create!(ThinArrow),
            '-' if consume!('-') => {
                let (token, recovered) = self.read_comment();
                let errors = recovered.then(|| TokenizerError {
                    error: TokenizerErrorType::UnclosedComment,
                    range: (start_position, self.source.position()),
                });
                self.create_recovered(start_position, token, errors.into_iter().collect())
            }
            '.' if matches!(self.source.current(), Some('0'..='9')) => {
                self.read_number(start_position, ".".to_string())
            }

            // Consume one
            ':' => create!(Colon),
            '+' => create!(Plus),
            '*' => create!(Star),
            '%' => create!(Percent),
            '^' => create!(Caret),
            '=' => create!(Equal),
            '~' => create!(Tilde),
            '<' => create!(LessThan),
            '>' => create!(GreaterThan),
            '(' => create!(LeftParen),
            ')' => create!(RightParen),
            ']' => create!(RightBracket),
            ';' => create!(Semicolon),
            '&' => create!(Ampersand),
            '|' => create!(Pipe),
            '?' => create!(QuestionMark),
            '#' => create!(Hash),
            ',' => create!(Comma),
            '[' => create!(LeftBracket),
            '}' => create!(RightBrace),
            '/' => create!(Slash),
            '.' => create!(Dot),
            '-' => create!(Minus),
            '{' => create!(LeftBrace),

            '\n' => Some(LexerResult::Ok(Token {
                token_type: TokenType::Whitespace { characters: ShortString::from("\n") },
                start_position,
                end_position: Position { bytes: start_position.bytes() + 1, ..start_position },
            })),

            // Oupsy
            unknown_char => Some(LexerResult::Fatal(vec![TokenizerError {
                error: TokenizerErrorType::UnexpectedToken(unknown_char),
                range: (start_position, self.source.position()),
            }])),
        }
    }

    fn read_number(&mut self, start_position: Position, mut number: String) -> Option<LexerResult<Token>> {
        let mut hit_decimal = false;
        while let Some(next) = self.source.current() {
            if next.is_ascii_digit() || matches!(next, '_') {
                number.push(self.source.next().expect("peeked, but no next"));
            } else if matches!(next, '.') && hit_decimal {
                return Some(self.eat_invalid_number(start_position, number));
            } else if matches!(next, '.') {
                hit_decimal = true;
                number.push(self.source.next().expect("peeked, but no next"));
            } else if matches!(next, 'e' | 'E') {
                return self.read_exponent_part(start_position, number);
            } else {
                break;
            }
        }
        self.create(start_position, TokenType::Number { text: ShortString::from(number) })
    }

    fn eat_invalid_number(&mut self, start_position: Position, mut number: String) -> LexerResult<Token> {
        loop {
            if matches!(self.source.current(), Some(token) if token.is_ascii_whitespace())
                || self.source.current().is_none()
            {
                return LexerResult::new(
                    Token {
                        token_type: TokenType::Number { text: number.into() },
                        start_position,
                        end_position: self.source.position(),
                    },
                    vec![TokenizerError {
                        error: TokenizerErrorType::InvalidNumber,
                        range: (start_position, self.source.position()),
                    }],
                );
            }

            number.push(self.source.next().expect("peeked, but no next"));
        }
    }

    // Starts from the exponent marker (like 'e')
    fn read_exponent_part(&mut self, start_position: Position, mut number: String) -> Option<LexerResult<Token>> {
        number.push(self.source.next().expect("peeked, but no next"));

        let next = self.source.current();
        if matches!(next, Some('+') | Some('-')) {
            number.push(self.source.next().expect("peeked, but no next"));
        }

        if !matches!(self.source.current(), Some('0'..='9')) {
            return Some(self.eat_invalid_number(start_position, number));
        }

        while let Some(next) = self.source.current() {
            if !next.is_ascii_digit() && !matches!(next, '_') {
                break;
            }
            number.push(self.source.next().expect("peeked, but no next"));
        }

        self.create(start_position, TokenType::Number { text: ShortString::from(number) })
    }

    fn read_hex_number(&mut self, hex_character: char, start_position: Position) -> Option<LexerResult<Token>> {
        let mut number = String::from_iter(['0', hex_character]);
        let mut hit_decimal = false;

        while let Some(next) = self.source.current() {
            match next {
                '0'..='9' | 'a'..='f' | 'A'..='F' | '_' => {
                    number.push(self.source.next().expect("peeked, but no next"))
                }

                'p' | 'P' if number.len() == 2 => return Some(self.eat_invalid_number(start_position, number)),
                'p' | 'P' => return self.read_exponent_part(start_position, number),
                '.' if hit_decimal => return Some(self.eat_invalid_number(start_position, number)),

                '.' => {
                    hit_decimal = true;
                    number.push(self.source.next().expect("peeked, but no next"));
                }

                _ => break,
            }
        }

        if number.len() == 2 {
            return Some(self.eat_invalid_number(start_position, number));
        }

        self.create(start_position, TokenType::Number { text: ShortString::from(number) })
    }

    fn read_binary_number(&mut self, binary_character: char, start_position: Position) -> Option<LexerResult<Token>> {
        let mut number = String::from_iter(['0', binary_character]);

        while matches!(self.source.current(), Some('0' | '1' | '_')) {
            number.push(self.source.next().expect("peeked, but no next"))
        }

        if number.len() == 2 {
            return Some(self.eat_invalid_number(start_position, number));
        }

        self.create(start_position, TokenType::Number { text: ShortString::from(number) })
    }

    // (string, had to be recovered?)
    fn read_string(&mut self, quote: char) -> (TokenType, bool) {
        let quote_type = match quote {
            '"' => StringLiteralQuoteType::Double,
            '\'' => StringLiteralQuoteType::Single,
            _ => unreachable!(),
        };

        let mut literal = String::new();
        let mut escape = false;
        let mut z_escaped = false;

        loop {
            let Some(next) = self.source.next() else {
                let token = TokenType::StringLiteral { literal: literal.into(), multi_line_depth: 0, quote_type };
                return (token, true);
            };

            match (escape, next) {
                (true, 'z') => {
                    escape = false;
                    z_escaped = true;
                    literal.push('z');
                }

                (true, ..) => {
                    escape = false;
                    z_escaped = true; // support for '\' followed by a new line
                    literal.push(next);
                }

                (false, '\\') => {
                    escape = true;
                    literal.push('\\');
                }

                (false, '\n' | '\r') if z_escaped => {
                    z_escaped = false;
                    literal.push(next);
                }

                (false, '\n' | '\r') => {
                    let token = TokenType::StringLiteral { literal: literal.into(), multi_line_depth: 0, quote_type };
                    return (token, true);
                }

                (false, ..) if next == quote => {
                    let token = TokenType::StringLiteral { literal: literal.into(), multi_line_depth: 0, quote_type };
                    return (token, false);
                }

                (false, ..) => literal.push(next),
            }
        }
    }

    // (comment, had to be recovered?)
    fn read_comment(&mut self) -> (TokenType, bool) {
        let mut comment = String::new();

        if self.source.consume('[') {
            match self.read_multi_line_body() {
                MultiLineBodyResult::Ok { blocks, body } | MultiLineBodyResult::Unclosed { blocks, body } => {
                    return (TokenType::MultiLineComment { blocks, comment: body.into() }, false);
                }

                MultiLineBodyResult::NotMultiLine { blocks } => {
                    comment.push('[');
                    (0..blocks).for_each(|_| comment.push('='));
                }
            }
        }

        let mut position_before_new_line = self.source.lexer_position;

        while let Some(next) = self.source.next() {
            if next == '\n' {
                break;
            }
            comment.push(next);
            position_before_new_line = self.source.lexer_position;
        }

        self.source.lexer_position = position_before_new_line;
        (TokenType::SingleLineComment { comment: comment.into() }, false)
    }

    fn read_multi_line_body(&mut self) -> MultiLineBodyResult {
        let mut blocks = 0;
        while self.source.consume('=') {
            blocks += 1;
        }

        if !self.source.consume('[') {
            return MultiLineBodyResult::NotMultiLine { blocks };
        }

        let mut body = String::new();
        'read_comment_char: loop {
            let Some(next) = self.source.next() else {
                return MultiLineBodyResult::Unclosed { blocks, body };
            };

            if next == ']' {
                let mut equal_signs = 0;
                while equal_signs < blocks {
                    if !self.source.consume('=') {
                        body.push(']');
                        (0..equal_signs).for_each(|_| body.push('='));
                        continue 'read_comment_char;
                    }
                    equal_signs += 1;
                }

                if self.source.consume(']') {
                    break;
                }

                body.push(']');
                (0..equal_signs).for_each(|_| body.push('='));

                continue;
            }

            body.push(next);
        }

        MultiLineBodyResult::Ok { blocks, body }
    }
}

fn is_identifier_start(character: char) -> bool {
    matches!(character, 'a'..='z' | 'A'..='Z' | '_')
}

pub(crate) struct LexerSource {
    source: Vec<char>,
    lexer_position: LexerPosition,
}

impl LexerSource {
    fn new(source: &str) -> Self {
        Self { source:  source.chars().collect(), lexer_position: LexerPosition::new() }
    }

    pub(crate) fn current(&self) -> Option<char> {
        self.source.get(self.lexer_position.index).copied()
    }

    pub(crate) fn next(&mut self) -> Option<char> {
        let next = self.current()?;
        if next == '\n' {
            self.lexer_position.position.line += 1;
            self.lexer_position.position.character = 1;
        } else {
            self.lexer_position.position.character += 1;
        }
        self.lexer_position.position.bytes += next.len_utf8();
        self.lexer_position.index += 1;
        Some(next)
    }

    pub(crate) fn peek(&self) -> Option<char> {
        self.source.get(self.lexer_position.index + 1).copied()
    }

    pub(crate) fn consume(&mut self, character: char) -> bool {
        if self.current() == Some(character) {
            self.next();
            return true;
        }
        false
    }

    pub(crate) fn consume_two(&mut self, one: char, two: char) -> bool {
        if self.current() == Some(one) && self.peek() == Some(two) {
            self.next();
            self.next();
            return true;
        }
        false
    }

    pub(crate) fn position(&self) -> Position {
        self.lexer_position.position
    }
}

#[derive(Clone, Copy)]
struct LexerPosition {
    position: Position,
    index: usize,
}

impl LexerPosition {
    const fn new() -> Self {
        Self { position: Position { line: 1, character: 1, bytes: 0 }, index: 0 }
    }
}

/// The result of a lexer operation.
#[derive(Debug, Deserialize, Serialize)]
pub enum LexerResult<T> {
    /// The lexer operation was successful.
    Ok(T),

    /// The lexer operation was unsuccessful, and could not be recovered.
    Fatal(Vec<TokenizerError>),

    /// The lexer operation was unsuccessful, but some result can be extracted.
    Recovered(T, Vec<TokenizerError>),
}

impl<T: std::fmt::Debug> LexerResult<T> {
    fn new(value: T, errors: Vec<TokenizerError>) -> Self {
        if errors.is_empty() {
            Self::Ok(value)
        } else {
            Self::Recovered(value, errors)
        }
    }

    /// Unwraps the result, panicking if it is not [`LexerResult::Ok`].
    #[allow(dead_code)]
    pub fn unwrap(self) -> T {
        match self {
            Self::Ok(value) => value,
            _ => panic!("expected ok, got {self:#?}"),
        }
    }

    /// Unwraps the errors, panicking if it is [`LexerResult::Ok`].
    pub fn unwrap_errors(self) -> Vec<TokenizerError> {
        match self {
            Self::Fatal(errors) | Self::Recovered(_, errors) => errors,
            _ => panic!("expected fatal error, got {self:#?}"),
        }
    }

    /// Returns the errors, if there was any.
    #[allow(dead_code)]
    pub fn errors(self) -> Vec<TokenizerError> {
        match self {
            Self::Recovered(_, errors) => errors,
            _ => Vec::new(),
        }
    }
}

enum MultiLineBodyResult {
    Ok { blocks: usize, body: String },
    NotMultiLine { blocks: usize },
    Unclosed { blocks: usize, body: String },
}
