use crate::{
    tokenizer::{Lexer, LexerResult},
    visitors::{Visit, VisitMut, Visitor, VisitorMut},
    ShortString,
};
use derive_more::{Deref, Display};
use serde::{Deserialize, Serialize};
use std::{cmp::Ordering, fmt};

macro_rules! symbol {
    {
        $(#[$symbol_meta:meta])*
        pub enum Symbol {
            $(
                $(#[$meta:meta])*
                $name:ident => $string:literal,
            )+
        }
    } => {
        paste::paste! {
            $(#[$symbol_meta])*
            #[derive(Display)]
            pub enum Symbol {
                $(
                    $(#[$meta])*
                    #[serde(rename = $string)]
                    #[display("{}", $string)]
                    $name,
                )+
            }

            impl Symbol {
                /// Given just the symbol text (no whitespace), returns the associated symbol, if
                /// it exists. If you want a TokenReference instead, consider [TokenReference::symbol].
                #[allow(clippy::should_implement_trait)]
                pub fn from_str(symbol: &str) -> Option<Self> {
                    match symbol {
                        $($string => { Some(Self::$name) },)+
                        _ => None,
                    }
                }

                /// Same as [Symbol::from_str], but can be called at compile time.
                pub const fn from_const_str(symbol: &'static str) -> Option<Self> {
                    pub const fn equal(lhs: &[u8], rhs: &[u8]) -> bool {
                        lhs.len() == rhs.len() && {
                            let (mut i, mut ret) = (0, true);
                            while i < lhs.len() {
                                ret &= lhs[i] == rhs[i];
                                i += 1;
                            }
                            ret
                        }
                    }
                    $({ if equal(symbol.as_bytes(), $string.as_bytes()) { return Some(Self::$name) } })+
                    None
                }
            }

        }
    };
}

symbol! {
    #[non_exhaustive]
    #[allow(missing_docs)]
    #[derive(Debug, Clone, Copy, PartialEq, Eq, Deserialize, Serialize)]
    /// A literal symbol, used for both words important to syntax (like while) and operators (like +)
    pub enum Symbol {
        And      => "and",
        Break    => "break",
        Do       => "do",
        Else     => "else",
        ElseIf   => "elseif",
        ElIf     => "elif",
        End      => "end",
        False    => "false",
        For      => "for",
        Function => "function",
        If       => "if",
        In       => "in",
        Local    => "local",
        Nil      => "nil",
        Not      => "not",
        Or       => "or",
        Repeat   => "repeat",
        Return   => "return",
        Then     => "then",
        True     => "true",
        Until    => "until",
        While    => "while",
        Import   => "import",
        Job      => "job",
        Option   => "option",
        Main     => "main",
        Yield    => "yield",

        PlusEqual        => "+=",
        MinusEqual       => "-=",
        StarEqual        => "*=",
        SlashEqual       => "/=",
        DoubleSlashEqual => "//=",
        PercentEqual     => "%=",
        CaretEqual       => "^=",
        TwoDotsEqual     => "..=",

        Ampersand         => "&",
        ThinArrow         => "->",
        TwoColons         => "::",
        Caret             => "^",
        Colon             => ":",
        Comma             => ",",
        Dot               => ".",
        TwoDots           => "..",
        Equal             => "=",
        TwoEqual          => "==",
        GreaterThan       => ">",
        GreaterThanEqual  => ">=",
        Hash              => "#",
        LeftBrace         => "{",
        LeftBracket       => "[",
        LeftParen         => "(",
        LessThan          => "<",
        LessThanEqual     => "<=",
        Minus             => "-",
        Percent           => "%",
        Pipe              => "|",
        Plus              => "+",
        QuestionMark      => "?",
        RightBrace        => "}",
        RightBracket      => "]",
        RightParen        => ")",
        Semicolon         => ";",
        Slash             => "/",
        DoubleSlash       => "//",
        Star              => "*",
        Tilde             => "~",
        TildeEqual        => "~=",
        DoubleGreaterThan => "~>",
        DoubleLesserThan  => "<~",
    }
}

/// The possible errors that can happen while tokenizing.
#[derive(Clone, Debug, PartialEq, Eq, Display, Deserialize, Serialize)]
pub enum TokenizerErrorType {
    /// An unclosed multi-line comment was found
    #[display("unclosed comment")]
    UnclosedComment,

    /// An unclosed string was found
    #[display("unclosed string")]
    UnclosedString,

    /// An invalid number was found
    #[display("invalid number")]
    InvalidNumber,

    /// An unexpected token was found
    #[display("unexpected character {_0}")]
    UnexpectedToken(char),

    /// Symbol passed is not valid
    /// Returned from [`TokenReference::symbol`]
    #[display("invalid symbol {_0}")]
    InvalidSymbol(String),
}

/// Used by serde
fn is_usize_zero(input: &usize) -> bool {
    *input == 0
}

/// The type of tokens in parsed code
#[derive(Clone, Debug, Eq, PartialEq, Deserialize, Serialize)]
#[serde(tag = "type")]
#[non_exhaustive]
pub enum TokenType {
    /// End of file, should always be the very last token
    Eof,

    /// An identifier, such as `foo`
    Identifier {
        /// The identifier itself
        identifier: ShortString,
    },

    /// A multi line comment in the format of `--[[ comment ]]`
    MultiLineComment {
        /// Number of equals signs, if any, for the multi line comment
        /// For example, `--[=[` would have a `blocks` value of `1`
        blocks: usize,

        /// The comment itself, ignoring opening and closing tags
        comment: ShortString,
    },

    /// A literal number, such as `3.3`
    Number {
        /// The text representing the number, includes details such as `0x`
        text: ShortString,
    },

    /// A shebang line
    Shebang {
        /// The shebang line itself
        line: ShortString,
    },

    /// A single line comment, such as `-- comment`
    SingleLineComment {
        /// The comment, ignoring initial `--`
        comment: ShortString,
    },

    /// A literal string, such as "Hello, world"
    StringLiteral {
        /// The literal itself, ignoring quotation marks
        literal: ShortString,

        #[serde(skip_serializing_if = "is_usize_zero")]
        /// Number of equals signs used for a multi line string, if it is one
        /// For example, `[=[string]=]` would have a `multi_line_depth` value of 1
        /// `[[string]]` would have a `multi_line_depth` value of 0
        /// A string such as `"string"` would have also have a `multi_line_depth` value of 0
        multi_line_depth: usize,

        /// The type of quotation mark used to make the string
        quote_type: StringLiteralQuoteType,
    },

    /// A [`Symbol`], such as `local` or `+`
    Symbol {
        /// The symbol itself
        symbol: Symbol,
    },

    /// Whitespace, such as tabs or new lines
    Whitespace {
        /// Characters consisting of the whitespace
        characters: ShortString,
    },
}

impl TokenType {
    /// Returns whether a token can be practically ignored in most cases
    /// Comments and whitespace will return `true`, everything else will return `false`
    pub fn is_trivia(&self) -> bool {
        use TokenType::*;
        matches!(self, Shebang { .. } | SingleLineComment { .. } | MultiLineComment { .. } | Whitespace { .. })
    }

    /// Returns the kind of the token type.
    ///
    /// ```rust
    /// use vvs_parser::prelude::ast::*;
    ///
    /// assert_eq!(
    ///     TokenType::Identifier {
    ///         identifier: ShortString::new("hello")
    ///     }.kind(),
    ///     TokenKind::Identifier,
    /// );
    /// ```
    pub const fn kind(&self) -> TokenKind {
        match self {
            TokenType::Eof => TokenKind::Eof,
            TokenType::Identifier { .. } => TokenKind::Identifier,
            TokenType::MultiLineComment { .. } => TokenKind::MultiLineComment,
            TokenType::Number { .. } => TokenKind::Number,
            TokenType::Shebang { .. } => TokenKind::Shebang,
            TokenType::SingleLineComment { .. } => TokenKind::SingleLineComment,
            TokenType::StringLiteral { .. } => TokenKind::StringLiteral,
            TokenType::Symbol { .. } => TokenKind::Symbol,
            TokenType::Whitespace { .. } => TokenKind::Whitespace,
        }
    }

    /// Returns a whitespace `TokenType` consisting of spaces
    pub fn spaces(spaces: usize) -> Self {
        TokenType::Whitespace { characters: " ".repeat(spaces).into() }
    }

    /// Returns a whitespace `TokenType` consisting of tabs
    pub fn tabs(tabs: usize) -> Self {
        TokenType::Whitespace { characters: "\t".repeat(tabs).into() }
    }

    /// Try to convert the token into a string slice. If it's not possible returns [None]. In the
    /// [TokenType::StringLiteral] case, we try to unquote the string.
    pub fn to_str(&self) -> Option<&str> {
        use StringLiteralQuoteType::*;
        match self {
            TokenType::Identifier { identifier } => Some(identifier),
            TokenType::StringLiteral { literal, quote_type: Double, .. } => Some(literal.trim_matches('\"').trim()),
            TokenType::StringLiteral { literal, quote_type: Single, .. } => Some(literal.trim_matches('\'').trim()),
            TokenType::StringLiteral { literal, quote_type: Brackets, multi_line_depth } => {
                todo!("use {multi_line_depth} to get ride of the [==[ or the [[ or the [=[ in {literal}")
            }
            _ => None,
        }
    }

    /// Same as [TokenType::to_str] but panics instead of returning [None].
    pub fn as_str(&self) -> &str {
        self.to_str().unwrap()
    }
}

impl fmt::Display for TokenType {
    fn fmt(&self, formatter: &mut fmt::Formatter) -> fmt::Result {
        match self {
            TokenType::Eof => Ok(()),
            TokenType::Number { text } => text.fmt(formatter),
            TokenType::Identifier { identifier } => identifier.fmt(formatter),
            TokenType::MultiLineComment { blocks, comment } => {
                write!(formatter, "--[{0}[{1}]{0}]", "=".repeat(*blocks), comment)
            }
            TokenType::Shebang { line } => line.fmt(formatter),
            TokenType::SingleLineComment { comment } => write!(formatter, "--{comment}"),
            TokenType::StringLiteral { literal, multi_line_depth, quote_type } => {
                if *quote_type == StringLiteralQuoteType::Brackets {
                    write!(formatter, "[{0}[{1}]{0}]", "=".repeat(*multi_line_depth), literal)
                } else {
                    write!(formatter, "{0}{1}{0}", quote_type.to_string(), literal)
                }
            }
            TokenType::Symbol { symbol } => symbol.fmt(formatter),
            TokenType::Whitespace { characters } => characters.fmt(formatter),
        }
    }
}

/// The kind of token. Contains no additional data.
#[derive(Clone, Copy, Debug, Eq, PartialEq)]
#[non_exhaustive]
pub enum TokenKind {
    /// End of file, should always be the very last token
    Eof,

    /// An identifier, such as `foo`
    Identifier,

    /// A multi line comment in the format of `--[[ comment ]]`
    MultiLineComment,

    /// A literal number, such as `3.3`
    Number,

    /// The shebang line
    Shebang,

    /// A single line comment, such as `-- comment`
    SingleLineComment,

    /// A literal string, such as "Hello, world"
    StringLiteral,

    /// A [`Symbol`], such as `local` or `+`
    Symbol,

    /// Whitespace, such as tabs or new lines
    Whitespace,
}

/// A token such consisting of its [`Position`] and a [`TokenType`]
#[derive(Clone, Debug, Eq, PartialEq, Display, Deserialize, Serialize)]
#[display("{token_type}")]
pub struct Token {
    pub(crate) start_position: Position,
    pub(crate) end_position: Position,
    pub(crate) token_type: TokenType,
}

impl Token {
    /// Creates a token with a zero position
    pub const fn new(token_type: TokenType) -> Token {
        Token { start_position: Position::new(), end_position: Position::new(), token_type }
    }

    /// The position a token begins at
    pub const fn start_position(&self) -> Position {
        self.start_position
    }

    /// The position a token ends at
    pub const fn end_position(&self) -> Position {
        self.end_position
    }

    /// The type of token as well as the data needed to represent it
    /// If you don't need any other information, use [`token_kind`](Token::token_kind) instead.
    pub const fn token_type(&self) -> &TokenType {
        &self.token_type
    }

    /// The kind of token with no additional data.
    /// If you need any information such as idenitfier names, use [`token_type`](Token::token_type) instead.
    pub const fn token_kind(&self) -> TokenKind {
        self.token_type().kind()
    }
}

impl Ord for Token {
    fn cmp(&self, other: &Self) -> Ordering {
        self.start_position().cmp(&other.start_position())
    }
}

impl PartialOrd for Token {
    fn partial_cmp(&self, other: &Self) -> Option<Ordering> {
        Some(self.cmp(other))
    }
}

impl Visit for Token {
    fn visit<V: Visitor>(&self, visitor: &mut V) {
        visitor.visit_token(self);
        match self.token_kind() {
            TokenKind::Eof => {}
            TokenKind::Identifier => visitor.visit_identifier(self),
            TokenKind::MultiLineComment => visitor.visit_multi_line_comment(self),
            TokenKind::Number => visitor.visit_number(self),
            TokenKind::Shebang => {}
            TokenKind::SingleLineComment => visitor.visit_single_line_comment(self),
            TokenKind::StringLiteral => visitor.visit_string_literal(self),
            TokenKind::Symbol => visitor.visit_symbol(self),
            TokenKind::Whitespace => visitor.visit_whitespace(self),
        }
    }
}

impl VisitMut for Token {
    fn visit_mut<V: VisitorMut>(self, visitor: &mut V) -> Self {
        let token = visitor.visit_token(self);
        match token.token_kind() {
            TokenKind::Eof => token,
            TokenKind::Identifier => visitor.visit_identifier(token),
            TokenKind::MultiLineComment => visitor.visit_multi_line_comment(token),
            TokenKind::Number => visitor.visit_number(token),
            TokenKind::Shebang => token,
            TokenKind::SingleLineComment => visitor.visit_single_line_comment(token),
            TokenKind::StringLiteral => visitor.visit_string_literal(token),
            TokenKind::Symbol => visitor.visit_symbol(token),
            TokenKind::Whitespace => visitor.visit_whitespace(token),
        }
    }
}

/// A reference to a token used by Ast's.
/// Dereferences to a [`Token`]
#[derive(Clone, Debug, Deref, Deserialize, Serialize)]
pub struct TokenReference {
    pub(crate) leading_trivia: Vec<Token>,
    #[deref]
    pub(crate) token: Token,
    pub(crate) trailing_trivia: Vec<Token>,
}

impl TokenReference {
    /// Creates a TokenReference from leading/trailing trivia as well as the leading token
    pub const fn new(leading_trivia: Vec<Token>, token: Token, trailing_trivia: Vec<Token>) -> Self {
        Self { leading_trivia, token, trailing_trivia }
    }

    /// Returns a symbol with the leading and trailing whitespace
    /// Only whitespace is supported
    /// ```rust
    /// # use vvs_parser::prelude::{ast::*, *};
    /// # fn main() -> Result<(), Box<TokenizerErrorType>> {
    /// let symbol = TokenReference::symbol("\nreturn ")?;
    /// assert_eq!(symbol.leading_trivia().next().unwrap().to_string(), "\n");
    /// assert_eq!(symbol.token().token_type(), &TokenType::Symbol { symbol: Symbol::Return });
    /// assert_eq!(symbol.trailing_trivia().next().unwrap().to_string(), " ");
    /// assert!(TokenReference::symbol("isnt whitespace").is_err());
    /// assert!(TokenReference::symbol(" notasymbol ").is_err());
    /// # Ok(())
    /// # }
    /// ```
    pub fn symbol(text: &str) -> Result<Self, TokenizerErrorType> {
        let mut lexer = Lexer::new_lazy(text);
        let mut leading_trivia = Vec::new();
        let symbol;
        loop {
            match lexer.process_next().expect("we shouldn't have hit eof") {
                LexerResult::Ok(token @ Token { token_type: TokenType::Whitespace { .. }, .. }) => {
                    leading_trivia.push(token)
                }
                LexerResult::Ok(token @ Token { token_type: TokenType::Symbol { .. }, .. }) => {
                    symbol = token;
                    break;
                }
                LexerResult::Ok(Token { token_type: TokenType::Eof, .. }) => {
                    return Err(TokenizerErrorType::InvalidSymbol(text.to_owned()))
                }
                LexerResult::Ok(token) => {
                    return Err(TokenizerErrorType::UnexpectedToken(token.to_string().chars().next().unwrap()))
                }
                LexerResult::Fatal(mut errors) | LexerResult::Recovered(_, mut errors) => {
                    return Err(errors.remove(0).error);
                }
            }
        }

        let mut trailing_trivia = Vec::new();
        loop {
            match lexer.process_next().expect("we shouldn't have hit eof") {
                LexerResult::Ok(token @ Token { token_type: TokenType::Whitespace { .. }, .. }) => {
                    trailing_trivia.push(token);
                }
                LexerResult::Ok(Token { token_type: TokenType::Eof, .. }) => break,
                LexerResult::Ok(token) => {
                    return Err(TokenizerErrorType::UnexpectedToken(token.to_string().chars().next().unwrap()));
                }
                LexerResult::Fatal(mut errors) | LexerResult::Recovered(_, mut errors) => {
                    return Err(errors.remove(0).error);
                }
            }
        }

        Ok(TokenReference { leading_trivia, token: symbol, trailing_trivia })
    }

    /// Create a symbol from its text representation. Will panic if the passed text is not a valid
    /// symbol...
    pub const fn basic_symbol(text: &'static str) -> Self {
        match Symbol::from_const_str(text) {
            None => panic!("unknown symbol"),
            Some(symbol) => TokenReference {
                leading_trivia: Vec::new(),
                token: Token::new(TokenType::Symbol { symbol }),
                trailing_trivia: Vec::new(),
            },
        }
    }

    /// Returns the inner token.
    pub const fn token(&self) -> &Token {
        &self.token
    }

    /// Returns the leading trivia
    pub fn leading_trivia(&self) -> impl Iterator<Item = &Token> {
        self.leading_trivia.iter()
    }

    /// Returns the trailing trivia
    pub fn trailing_trivia(&self) -> impl Iterator<Item = &Token> {
        self.trailing_trivia.iter()
    }

    /// Creates a clone of the current TokenReference with the new inner token, preserving trivia.
    pub fn with_token(&self, token: Token) -> Self {
        Self { token, leading_trivia: self.leading_trivia.clone(), trailing_trivia: self.trailing_trivia.clone() }
    }

    /// Checks if the token is the given symbol
    pub fn is_symbol(&self, symbol: Symbol) -> bool {
        self.token.token_type() == &TokenType::Symbol { symbol }
    }

    /// Checks if the token is in the given symbol set
    pub fn in_symbols(&self, symbols: &[Symbol]) -> bool {
        symbols.iter().any(|&symbol| self.is_symbol(symbol))
    }

    /// Checks if the token is of the given kind
    pub fn is_kind(&self, kind: TokenKind) -> bool {
        self.token.token_kind() == kind
    }
}

impl std::borrow::Borrow<Token> for &TokenReference {
    fn borrow(&self) -> &Token {
        self
    }
}

impl fmt::Display for TokenReference {
    fn fmt(&self, formatter: &mut fmt::Formatter) -> fmt::Result {
        for trivia in &self.leading_trivia {
            trivia.fmt(formatter)?;
        }
        self.token.fmt(formatter)?;
        for trivia in &self.trailing_trivia {
            trivia.fmt(formatter)?;
        }
        Ok(())
    }
}

impl From<&TokenReference> for TokenReference {
    fn from(value: &TokenReference) -> Self {
        value.clone()
    }
}

impl PartialEq<Self> for TokenReference {
    fn eq(&self, other: &Self) -> bool {
        (**self).eq(other)
            && self.leading_trivia == other.leading_trivia
            && self.trailing_trivia == other.trailing_trivia
    }
}

impl Eq for TokenReference {}

impl Ord for TokenReference {
    fn cmp(&self, other: &Self) -> Ordering {
        (**self).cmp(&**other)
    }
}

impl PartialOrd for TokenReference {
    fn partial_cmp(&self, other: &Self) -> Option<Ordering> {
        Some(self.cmp(other))
    }
}

impl Visit for TokenReference {
    fn visit<V: Visitor>(&self, visitor: &mut V) {
        visitor.visit_token(self);
        if matches!(self.token().token_kind(), TokenKind::Eof) {
            visitor.visit_eof(self);
        }
        self.leading_trivia.visit(visitor);
        self.token.visit(visitor);
        self.trailing_trivia.visit(visitor);
    }
}

impl VisitMut for TokenReference {
    fn visit_mut<V: VisitorMut>(self, visitor: &mut V) -> Self {
        let mut token_reference = visitor.visit_token_reference(self);
        if matches!(token_reference.token().token_kind(), TokenKind::Eof) {
            token_reference = visitor.visit_eof(token_reference);
        }
        token_reference.leading_trivia = token_reference.leading_trivia.visit_mut(visitor);
        token_reference.token = token_reference.token.visit_mut(visitor);
        token_reference.trailing_trivia = token_reference.trailing_trivia.visit_mut(visitor);
        token_reference
    }
}

/// Used to represent exact positions of tokens in code
#[derive(Clone, Copy, Debug, Default, PartialEq, Eq, Deserialize, Serialize)]
pub struct Position {
    pub(crate) bytes: usize,
    pub(crate) line: usize,
    pub(crate) character: usize,
}

impl Position {
    /// Create a new default position. Can be called in const context.
    pub const fn new() -> Self {
        Self { bytes: 0, line: 0, character: 0 }
    }

    /// How many bytes, ignoring lines, it would take to find this position
    pub fn bytes(self) -> usize {
        self.bytes
    }

    /// Index of the character on the line for this position
    pub fn character(self) -> usize {
        self.character
    }

    /// Line the position lies on
    pub fn line(self) -> usize {
        self.line
    }
}

impl Ord for Position {
    fn cmp(&self, other: &Self) -> Ordering {
        self.bytes.cmp(&other.bytes)
    }
}

impl PartialOrd for Position {
    fn partial_cmp(&self, other: &Self) -> Option<Ordering> {
        Some(self.cmp(other))
    }
}

/// The types of quotes used in a Lua string
#[derive(Clone, Copy, Debug, Eq, PartialEq, Deserialize, Serialize)]
#[non_exhaustive]
pub enum StringLiteralQuoteType {
    /// Strings formatted \[\[with brackets\]\]
    Brackets,

    /// Strings formatted "with double quotes"
    Double,

    /// Strings formatted 'with single quotes'
    Single,
}

impl fmt::Display for StringLiteralQuoteType {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        match *self {
            // Brackets cannot be properly displayed, as not only do they have
            // variable depth (`=`), but also they don't open the same as
            // they end, meaning this can't really be used for display purposes.
            StringLiteralQuoteType::Brackets => Err(fmt::Error),
            StringLiteralQuoteType::Double => write!(f, "\""),
            StringLiteralQuoteType::Single => write!(f, "'"),
        }
    }
}

/// Information about an error that occurs while tokenizing
#[derive(Clone, Debug, PartialEq, Eq, Display, Deserialize, Serialize)]
#[display("{error} ({}:{} to {}:{})",
    range.0.line, range.0.character, range.1.line, range.1.character
)]
pub struct TokenizerError {
    /// The type of error
    pub(crate) error: TokenizerErrorType,

    /// The range of the token that caused the error
    pub(crate) range: (Position, Position),
}

impl TokenizerError {
    /// The type of error
    pub fn error(&self) -> &TokenizerErrorType {
        &self.error
    }

    /// The position of the first token that caused the error
    pub fn position(&self) -> Position {
        self.range.0
    }

    /// The range of the token that caused the error
    pub fn range(&self) -> (Position, Position) {
        self.range
    }
}

impl std::error::Error for TokenizerError {}
