use std::{
    cmp::Ordering,
    hash::{Hash, Hasher},
    marker::PhantomData,
    mem::{align_of, size_of},
    ptr::NonNull,
};

/// A tagged pointer: a space-efficient representation of a pointer that smuggles an integer tag
/// into it.
///
/// `BITS` specifies how many bits are used for the tag. The alignment of `T` must be large enough
/// to store this many bits.
///
/// # Safety
/// The passed pointers, at construction time and each time you modify this struct, should all be
/// properly aligned, or your code may panic. The passed pointer must also be dereferenceable.
#[repr(transparent)]
pub struct TaggedPtr<T, const BITS: u8>(NonNull<u8>, PhantomData<NonNull<T>>);

impl<T, const BITS: u8> TaggedPtr<T, BITS> {
    const ASSERT: () = {
        assert!(size_of::<usize>() == size_of::<u64>(), "we only work with 64 bits architectures");
        assert!(align_of::<T>().is_power_of_two(), "expected alignment of `T` to be a power of 2");
        assert!(
            match size_of::<T>() {
                0 => true,
                n => n >= align_of::<T>(),
            },
            "expected size of non-ZST `T` to be at least alignment"
        );
        assert!(size_of::<T>() % align_of::<T>() == 0, "expected size of `T` to be multiple of alignment");
        assert!(1_usize.checked_shl(BITS as u32).is_some(), "`1 << BITS` doesn't fit in a `usize`");
        assert!(align_of::<T>().trailing_zeros() >= BITS as u32, "alignment of `T` must be at least `1 << BITS`");
    };

    const ALIGNMENT: usize = 1_usize.wrapping_shl(BITS as u32);
    const MASK: usize = Self::ALIGNMENT - 1;

    /// Creates a new tagged pointer. Only the lower `BITS` bits of `tag` are stored.
    ///
    /// If build in debug, a check is performed at compile time to ensure that the alignment of `T`
    /// is sufficient to smuggle the tag.
    #[allow(path_statements, clippy::no_effect)]
    #[inline]
    pub fn new(ptr: NonNull<T>, tag: usize) -> Self {
        Self::ASSERT; // Static assert!
        let ptr = ptr.as_ptr().cast::<u8>();

        let offset = ptr.align_offset(Self::ALIGNMENT);
        debug_assert!(offset != usize::MAX, "align offset failed");
        debug_assert!(offset & Self::MASK == 0, "`ptr` is not aligned enough");

        let ptr = ptr.wrapping_add(tag & Self::MASK);
        Self(unsafe { NonNull::new_unchecked(ptr) }, PhantomData)
    }

    /// Gets the pointer and tag stored by the tagged pointer.
    #[inline]
    pub fn get(self) -> (NonNull<T>, usize) {
        let ptr = self.0.as_ptr();
        let tag = Self::ALIGNMENT.wrapping_sub(ptr.align_offset(Self::ALIGNMENT)) & Self::MASK;
        let ptr = ptr.wrapping_sub(tag).cast::<T>();

        // SAFETY: ptr + 1 - 1 == ptr
        (unsafe { NonNull::new_unchecked(ptr) }, tag)
    }

    /// Get the address of the pointed thing as an integer. Here for moar reasons we only work with
    /// 64 bit architectures, thus the `usize == u64`.
    #[inline]
    pub fn addr(self) -> u64 {
        self.ptr().as_ptr() as u64
    }

    /// Gets the pointer stored by the tagged pointer, without the tag.
    #[inline]
    pub fn ptr(self) -> NonNull<T> {
        self.get().0
    }

    /// Gets the tag stored by the tagged pointer.
    #[inline]
    pub fn tag(self) -> usize {
        self.get().1
    }

    /// Sets the pointer without modifying the tag.
    #[inline]
    pub fn set_ptr(&mut self, ptr: NonNull<T>) {
        *self = Self::new(ptr, self.tag());
    }

    /// Sets the tag without modifying the pointer.
    #[inline]
    pub fn set_tag(&mut self, tag: usize) {
        *self = Self::new(self.ptr(), tag);
    }
}

impl<T, const BITS: u8> Copy for TaggedPtr<T, BITS> {}
impl<T, const BITS: u8> Clone for TaggedPtr<T, BITS> {
    fn clone(&self) -> Self {
        *self
    }
}

impl<T, const BITS: u8> Eq for TaggedPtr<T, BITS> {}
impl<T, const BITS: u8> PartialEq for TaggedPtr<T, BITS> {
    fn eq(&self, other: &Self) -> bool {
        self.0 == other.0
    }
}

impl<T, const BITS: u8> PartialOrd for TaggedPtr<T, BITS> {
    fn partial_cmp(&self, other: &Self) -> Option<Ordering> {
        Some(self.cmp(other))
    }
}
impl<T, const BITS: u8> Ord for TaggedPtr<T, BITS> {
    fn cmp(&self, other: &Self) -> Ordering {
        self.0.cmp(&other.0)
    }
}

impl<T, const BITS: u8> Hash for TaggedPtr<T, BITS> {
    fn hash<H: Hasher>(&self, state: &mut H) {
        self.0.hash(state);
    }
}

impl<T, const BITS: u8> std::fmt::Debug for TaggedPtr<T, BITS> {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        let (ptr, tag) = self.get();
        f.debug_struct("TaggedPtr")
            .field("ptr", &ptr)
            .field("tag", &tag)
            .finish()
    }
}

#[cfg(any(test, doctest))]
mod test {
    use crate::{
        types::{VVRTAny, VVRTType},
        TaggedPtr,
    };
    use core::{mem::size_of, ptr::NonNull};
    use hashbrown::HashMap;
    use std::mem::align_of;

    #[test]
    fn test_ptr_size() {
        assert_eq!(size_of::<usize>(), size_of::<u64>());
        assert_eq!(size_of::<*const u64>(), size_of::<u64>());
    }

    #[test]
    fn test_ptr_align() {
        assert_eq!(size_of::<NonNull<u64>>(), 8);
        assert_eq!(align_of::<NonNull<u64>>(), 8);
        assert_eq!(align_of::<NonNull<HashMap<Box<str>, VVRTAny>>>(), 8);
        assert_eq!(align_of::<NonNull<Vec<VVRTAny>>>(), 8);
    }

    #[test]
    fn test_type_sizes() {
        use std::mem::size_of;
        assert_eq!(size_of::<VVRTAny>(), 8);
        assert_eq!(size_of::<VVRTType>(), 1);
    }

    #[repr(align(2))]
    #[derive(Debug, Eq, PartialEq)]
    struct Align2(pub u16);

    #[repr(align(4))]
    #[derive(Debug, Eq, PartialEq)]
    struct Align4(pub u32);

    #[repr(align(8))]
    #[derive(Debug, Eq, PartialEq)]
    struct Align8(pub u64);

    #[test]
    #[allow(clippy::let_unit_value)]
    fn check_ptr_basic() {
        for i in 0..64 {
            let x = i as u8 * 3;
            let tp = TaggedPtr::<_, 0>::new((&x).into(), i);
            assert_eq!(tp.get(), ((&x).into(), 0));
            assert_eq!(*unsafe { tp.ptr().as_ref() }, x);

            let x = Align2(i as u16 * 5);
            let tp = TaggedPtr::<_, 1>::new((&x).into(), i);
            assert_eq!(tp.get(), ((&x).into(), i & 0b1));
            assert_eq!(*unsafe { tp.ptr().as_ref() }, x);

            let mut x = Align4(i as u32 * 7);
            let ptr = NonNull::from(&mut x);
            let tp = TaggedPtr::<_, 2>::new(ptr, i);
            assert_eq!(tp.get(), (ptr, i & 0b11));
            unsafe { tp.ptr().as_mut() }.0 += 1;
            assert_eq!(x.0, i as u32 * 7 + 1);

            let x = Align8(i as u64 * 11);
            let tp = TaggedPtr::<_, 3>::new((&x).into(), i);
            assert_eq!(tp.get(), ((&x).into(), i & 0b111));
            assert_eq!(*unsafe { tp.ptr().as_ref() }, x);
        }
    }

    #[test]
    #[allow(clippy::let_unit_value)]
    fn check_ptr_overaligned() {
        for i in 0..64 {
            let x = Align2(i as u16 * 3);
            let tp = TaggedPtr::<_, 0>::new((&x).into(), i);
            assert_eq!(tp.get(), ((&x).into(), 0));
            assert_eq!(*unsafe { tp.ptr().as_ref() }, x);

            let x = Align4(i as u32 * 5);
            let tp = TaggedPtr::<_, 1>::new((&x).into(), i);
            assert_eq!(tp.get(), ((&x).into(), i & 0b1));
            assert_eq!(*unsafe { tp.ptr().as_ref() }, x);

            let x = Align8(i as u64 * 7);
            let tp = TaggedPtr::<_, 2>::new((&x).into(), i);
            assert_eq!(tp.get(), ((&x).into(), i & 0b11));
            assert_eq!(*unsafe { tp.ptr().as_ref() }, x);
        }
    }

    #[test]
    #[allow(clippy::let_unit_value)]
    fn check_ptr_zst() {
        for i in 0..8 {
            let x = ();
            let tp = TaggedPtr::<_, 0>::new((&x).into(), i);
            assert_eq!(tp.get(), ((&x).into(), 0));
        }
    }

    #[test]
    fn check_ptr_size() {
        assert_eq!(size_of::<TaggedPtr<u64, 3>>(), size_of::<core::ptr::NonNull<u64>>());
        assert_eq!(size_of::<TaggedPtr<u64, 3>>(), size_of::<Option<TaggedPtr<u64, 3>>>());
    }

    #[test]
    fn check_ptr_not_entirely_dereferenceable() {
        #[repr(align(8))]
        #[allow(dead_code)]
        struct Type(u64, u64, u64);

        let a = Type(1, 2, 3);
        let mut tp = TaggedPtr::<_, 2>::new((&a).into(), 0b10);
        assert_eq!(tp.get(), ((&a).into(), 0b10));

        let mut b = Align4(0);
        let ptr = NonNull::from(&mut b).cast();
        tp.set_ptr(ptr);
        assert_eq!(tp.ptr(), ptr);
        tp.set_tag(0b11);
        assert_eq!(tp.tag(), 0b11);
        unsafe { tp.ptr().cast::<Align4>().as_mut() }.0 = 1234;
        assert_eq!(b.0, 1234);
    }

    #[test]
    #[should_panic]
    fn runtime_ptr_not_aligned_enough() {
        let ptr: *mut Align2 = &mut Align2(0);
        let ptr = unsafe { (ptr as *mut u8).add(1) };
        let ptr = ptr as *mut Align2;
        let ptr = unsafe { NonNull::new_unchecked(ptr) };
        TaggedPtr::<_, 1>::new(ptr, 0);
    }
}
