use crate::types::*;
use hashbrown::HashMap;
use std::{ptr::NonNull, rc::Rc};

/// Aux struct used to store infos about runtime types that imbark a VVRTTable and where we don't
/// have enaugh space in the trailing zeros of the pointer to store them...
#[derive(Default, Debug, Clone, Copy, PartialEq, Eq)]
pub struct VVRTInfos {
    pub is_ass_borrowed: bool,
}

/// A table, really just a hashmap.
///
/// NOTE: Needs to be clone to be able to be used in the ASS things...
///
/// In a table we also embark an AUX structure to store runtime informations about ass types that
/// owns this table, or any other thing that may be necessary latter down the line... When
/// comparing tables we don't compare the [VVRTInfos] struct.
#[derive(Debug, Clone, Copy)]
#[repr(transparent)]
pub struct VVRTTable(NonNull<(HashMap<VVRTString, VVRTAny>, VVRTInfos)>);
crate::types::any::impl_try_from_any! { Table => VVRTTable }

impl VVRTTable {
    #[inline]
    pub fn dangling() -> Self {
        Self(NonNull::dangling())
    }

    #[inline]
    pub fn as_ptr(self) -> *mut u64 {
        self.0.as_ptr() as *mut _
    }

    #[inline]
    pub fn iter(&self) -> hashbrown::hash_map::Iter<VVRTString, VVRTAny> {
        unsafe { self.0.as_ref() }.0.iter()
    }

    #[inline]
    pub fn iter_mut(&mut self) -> hashbrown::hash_map::IterMut<VVRTString, VVRTAny> {
        unsafe { self.0.as_mut() }.0.iter_mut()
    }

    #[inline]
    pub fn rt_infos(&self) -> &VVRTInfos {
        &unsafe { self.0.as_ref() }.1
    }

    #[inline]
    pub fn rt_infos_mut(&mut self) -> &mut VVRTInfos {
        &mut unsafe { self.0.as_mut() }.1
    }
}

impl AsMut<HashMap<VVRTString, VVRTAny>> for VVRTTable {
    fn as_mut(&mut self) -> &mut HashMap<VVRTString, VVRTAny> {
        &mut unsafe { self.0.as_mut() }.0
    }
}

impl AsRef<HashMap<VVRTString, VVRTAny>> for VVRTTable {
    fn as_ref(&self) -> &HashMap<VVRTString, VVRTAny> {
        &unsafe { self.0.as_ref() }.0
    }
}

impl vvs_ass::AuxTable for VVRTTable {
    #[inline]
    fn deep_clone(&self) -> Self {
        unsafe { VVRTTable_deep_clone(*self) }
    }
}

impl Default for VVRTTable {
    #[inline]
    fn default() -> Self {
        HashMap::<VVRTString, VVRTAny>::default().into()
    }
}

impl From<HashMap<VVRTString, VVRTAny>> for VVRTTable {
    #[inline]
    fn from(value: HashMap<VVRTString, VVRTAny>) -> Self {
        Self(unsafe { NonNull::new_unchecked(Rc::into_raw(Rc::new(value)) as *mut _) })
    }
}

impl Eq for VVRTTable {}
impl PartialEq for VVRTTable {
    #[inline]
    fn eq(&self, other: &Self) -> bool {
        self.as_ref().eq(other.as_ref())
    }
}

/// Module used to define functions to be exported for the JIT.
#[allow(clippy::missing_safety_doc)]
#[vvs_procmacro::vvrt_bridge(VVRTTable)]
mod ffi {
    #[no_mangle]
    pub unsafe extern "C" fn new() -> VVRTTable {
        Default::default()
    }

    /// Deeply clone the table.
    #[no_mangle]
    pub unsafe extern "C" fn deep_clone(this: VVRTTable) -> VVRTTable {
        this.as_ref()
            .iter()
            .map(|(k, v)| (VVRTString_clone(*k), VVRTAny_deep_clone(*v)))
            .collect::<HashMap<_, _>>()
            .into()
    }

    /// Increment the count of each child because we own them. This function is not-reentrent and
    /// we don't support cyclic references...
    #[no_mangle]
    pub unsafe extern "C" fn clone(this @ VVRTTable(ptr): VVRTTable) -> VVRTTable {
        this.iter().for_each(|(key, value)| {
            let _ = VVRTString_clone(*key);
            let _ = VVRTAny_clone(*value);
        });
        Rc::increment_strong_count(ptr.as_ptr());
        VVRTTable(ptr)
    }

    // Decrement the count of each child because we own them. We don't support cyclic references...
    #[no_mangle]
    pub unsafe extern "C" fn drop(this @ VVRTTable(ptr): VVRTTable) {
        this.iter().for_each(|(key, value)| {
            VVRTString_drop(*key);
            VVRTAny_drop(*value);
        });
        Rc::decrement_strong_count(ptr.as_ptr())
    }

    /// Insert something into the table. Any old present value will be dropped. The key and value
    /// are consumed by this function.
    #[no_mangle]
    pub unsafe extern "C" fn insert(mut table: VVRTTable, key: VVRTString, value: VVRTAny) {
        debug_assert!(!matches!(value.ty(), VVRTType::Table), "can't store container in containers");
        if let Some(old) = table.as_mut().insert(key, value) {
            VVRTAny_drop(old)
        }
    }

    /// Insert something into the table. Any old present value will be dropped. The value is
    /// consumed by this function. The key string will be used to create a new string.
    #[no_mangle]
    pub unsafe extern "C" fn insert_from_raw_key(table: VVRTTable, key: *const u8, len: i32, value: VVRTAny) {
        let key = VVRTString_new_from_raw(key, len);
        VVRTTable_insert(table, key, value);
    }

    /// Get the content of of the table. The thing is cloned when getted from the table.
    #[no_mangle]
    pub unsafe extern "C" fn get(table: VVRTTable, key: VVRTString) -> VVRTAny {
        table
            .as_ref()
            .get(&key)
            .map(|item| VVRTAny_clone(*item))
            .unwrap_or_default()
    }

    #[rustfmt::skip] #[no_mangle] unsafe extern "C" fn eq(this: VVRTTable, other: VVRTTable) -> bool { PartialEq::eq(&this, &other) }
    #[rustfmt::skip] #[no_mangle] unsafe extern "C" fn ne(this: VVRTTable, other: VVRTTable) -> bool { PartialEq::ne(&this, &other) }
}
