use crate::types::*;
use anyhow::{anyhow, bail, ensure};
use std::{
    alloc,
    mem::{size_of, transmute},
    ptr::NonNull,
    slice, str,
};

/// Store all the registered variants. We share variants between colors and movements.
static mut VARIANTS: Vec<VariantDescription> = Vec::new();

struct VariantDescription {
    /// The name of the variant family in the AST, like color or movement.
    ast_name: &'static str,

    /// The name of the variant in its family, like `rgb`, `move`, etc.
    variant_name: &'static str,

    /// The content of this variant, we store the runtime types here.
    arguments: Vec<VVRTType>,

    /// The layout of this variant. If we have a string we don't store the content of said string
    /// here, we only take into account the integer used to store the length of the string.
    layout: alloc::Layout,

    /// The layouts of the content of this variant. Note that we also store the offset of the
    /// discriminent in this vector. Because of that, we have the following:
    /// `assert!(arguments.len() == offsets.len() + 1)`;
    offsets: Vec<usize>,
}

unsafe fn get_variant_argument_count_by_discriminent(discriminent: usize) -> i32 {
    VARIANTS[discriminent].arguments.len() as i32
}

unsafe fn get_variant_desc_by_name(
    ast: &str,
    var: &str,
) -> (usize, &'static alloc::Layout, &'static [usize], &'static [VVRTType]) {
    VARIANTS
        .iter()
        .enumerate()
        .find_map(
            |(idx, VariantDescription { ast_name, variant_name, layout, offsets, arguments })| {
                (*ast_name == ast && *variant_name == var).then_some((idx, layout, &**offsets, &**arguments))
            },
        )
        .expect("asked to build an invalid variant")
}

#[derive(Debug, Clone, Copy)]
#[repr(transparent)]
pub struct VVRTVariant(NonNull<u64>);
crate::types::any::impl_try_from_any! { Variant => VVRTVariant }

impl VVRTVariant {
    #[inline]
    pub fn dangling() -> Self {
        Self(NonNull::dangling())
    }

    #[inline]
    pub fn discriminent(&self) -> usize {
        unsafe { *transmute::<*mut u64, *const usize>(self.0.as_ptr()) }
    }

    /// # Safety
    /// This function is not thread safe.
    pub unsafe fn register_runtime_variant(
        ast_name: &'static str,
        variant_name: &'static str,
        arguments: Vec<VVRTType>,
    ) -> anyhow::Result<()> {
        use VVRTType::*;
        let name = format!("{ast_name}.{variant_name}");

        if let [heads @ .., last] = &arguments[..] {
            heads.iter().copied().enumerate().try_for_each(|(idx, ty)| {
                (!matches!(ty, Number))
                    .then_some(())
                    .ok_or_else(|| anyhow!("found invalid argument type `{ty}` at position {idx} in variant `{name}`"))
            })?;
            ensure!(
                matches!(last, String | Number),
                "found invalid argument type `{last}` at last position in variant `{name}`"
            );
        }

        (!VARIANTS
            .iter()
            .any(|desc| desc.ast_name == ast_name && desc.variant_name == variant_name))
        .then(|| -> anyhow::Result<()> {
            let (layout, offsets) = VVRTVariant::layout(&arguments)?;
            VARIANTS.push(VariantDescription { ast_name, variant_name, arguments, layout, offsets });
            Ok(())
        })
        .ok_or_else(|| anyhow!("redefinition of variant {name}"))?
    }

    /// Get the location of the variant.
    #[inline]
    pub fn as_ptr(self) -> *mut u64 {
        self.0.as_ptr()
    }

    /// Get the base of the memory corresponding to the passed index along side the type of this
    /// memory location.
    unsafe fn offset_of_index(&self, idx: i32) -> (*const u8, VVRTType) {
        let VariantDescription { arguments, offsets, .. } =
            VARIANTS.get(self.discriminent()).expect("variant was not valid");

        debug_assert!(arguments.len() == offsets.len() + 1, "invalid variant");
        debug_assert!(arguments.len() > idx as usize, "invalid index");

        // In the offsets we store the initial index for the discriminent in the offsets also!
        let ptr: *const u8 = self.0.as_ptr().byte_add(offsets[idx as usize + 1]) as *const _;
        (ptr, arguments[idx as usize])
    }

    /// Compute the layout for the content of a variant. Because of how we represent strings in a
    /// variant, a vairant is immutable, like strings. Here we don't put the space for the content
    /// of the string because we can't know it.
    fn layout(body: &[VVRTType]) -> anyhow::Result<(alloc::Layout, Vec<usize>)> {
        // We need the first element to be the discriminent of the variant. The discriminent is
        // the index in the variant vector, thus we need a usize.
        let init = (alloc::Layout::new::<usize>(), vec![0]);

        // We fold the body into the layout.
        let fold = |(body, mut offsets): (alloc::Layout, Vec<usize>), elem: VVRTType| {
            let (body, next) = body.extend(match elem {
                // Only the length. Because of the alignement, the next two bytes are always
                // allocated for the content of the string, so we don't waste too much space.
                VVRTType::String => alloc::Layout::new::<i32>(),

                // For numbers we directly use an any.
                VVRTType::Number => alloc::Layout::new::<VVRTAny>(),

                _ => bail!("invalid type `{elem}` in variant"),
            })?;
            offsets.push(next);
            Ok::<_, anyhow::Error>((body, offsets))
        };

        body.iter()
            .copied()
            .try_fold(init, fold)
            .map(|(layout, offsets)| (layout.pad_to_align(), offsets))
    }

    /// Get the layout for the current value. In this layout we have the space for the content of
    /// the string.
    fn get_value_layout(&self) -> alloc::Layout {
        let VariantDescription { arguments, layout, offsets, .. } = unsafe { &VARIANTS[self.discriminent()] };

        match (arguments.last(), offsets.last()) {
            (Some(VVRTType::String), Some(offset)) => {
                // Get the memory layout for where we store the string.
                let str_len = unsafe { *(self.0.as_ptr().add(*offset) as *const i32) as usize };
                let str_layout = alloc::Layout::array::<u8>(str_len);

                // Get the total layout. Just assert we are not doing things too big.
                let layout = layout.extend(str_layout.expect("too big")).expect("too big").0;

                // Because of C things, we must pad the layout up to the align value, which should
                // be the same as `*mut u64`.
                layout.pad_to_align()
            }

            _ => *layout,
        }
    }

    /// Like [VVRTVariant_get] but for [VVRTNumber]. Return default on type error.
    fn get_number(&self, idx: i32) -> VVRTNumber {
        match unsafe { self.offset_of_index(idx) } {
            (ptr, VVRTType::Number) => Into::into(unsafe { *(ptr as *const VVRTAny) }),
            _ => Default::default(),
        }
    }

    /// Like [self::VVRTVariant_get] but for the string in the variant. Return default on type
    /// error. Here we don't have an index because we know that we have only one string in the
    /// variant and that it's at the last index.
    fn get_string(&self) -> &str {
        match unsafe { self.offset_of_index(get_variant_argument_count_by_discriminent(self.discriminent()) - 1) } {
            (ptr, VVRTType::String) => unsafe {
                str::from_utf8_unchecked(slice::from_raw_parts(
                    ptr.byte_add(size_of::<i32>()),
                    *(ptr as *const i32) as usize,
                ))
            },
            _ => Default::default(),
        }
    }
}

impl AsRef<[u8]> for VVRTVariant {
    #[inline]
    fn as_ref(&self) -> &[u8] {
        unsafe { slice::from_raw_parts(self.0.as_ptr() as *const u8, self.get_value_layout().size()) }
    }
}

impl Eq for VVRTVariant {}
impl PartialEq for VVRTVariant {
    #[inline]
    fn eq(&self, other: &Self) -> bool {
        (self.discriminent() == other.discriminent()) && (self.as_ref() == other.as_ref())
    }
}

/// Module used to defined functions to be exported to the JIT.
#[allow(clippy::missing_safety_doc)]
#[vvs_procmacro::vvrt_bridge(VVRTVariant)]
mod ffi {
    use std::borrow::Cow;

    #[no_mangle]
    unsafe extern "C" fn new(
        ast_ptr: *const u8,
        ast_len: i32,
        var_ptr: *const u8,
        var_len: i32,
        str_ptr: *const u8,
        str_len: i32,
    ) -> VVRTVariant {
        let ast = str::from_utf8_unchecked(slice::from_raw_parts(ast_ptr, ast_len as usize));
        let var = str::from_utf8_unchecked(slice::from_raw_parts(var_ptr, var_len as usize));
        let (discriminent, layout, offsets, types) = get_variant_desc_by_name(ast, var);

        // Get the layout for the variant to allocate.
        let (layout, string_offset) = match types {
            [body @ .., _] if body.iter().copied().any(|t| t == VVRTType::String) => {
                unreachable!("strings must be in last position in variants")
            }
            [.., VVRTType::String] => {
                let (layout, next) = layout
                    .extend(alloc::Layout::array::<u8>(str_len as usize).expect("too big"))
                    .expect("too big");
                (layout.pad_to_align(), Some(next))
            }
            _ => (*layout, None),
        };

        // Create and initialize the variant. Pay attention that the unit of the pointer is a byte
        // (a u8...)
        let ptr: NonNull<u8> = NonNull::new(alloc::alloc(layout)).expect("out of memory");
        *transmute::<*mut u8, *mut usize>(ptr.as_ptr()) = discriminent;
        slice::from_raw_parts_mut(ptr.as_ptr(), layout.size())
            .iter_mut()
            .skip(size_of::<usize>())
            .for_each(|byte: &mut u8| std::ptr::write(byte, 0));

        // If we have a string we must copy it.
        if let Some(string_offset) = string_offset {
            std::ptr::write(
                ptr.as_ptr().byte_add(*offsets.last().expect("invalid variant")) as *mut i32,
                str_len,
            );
            std::ptr::copy_nonoverlapping(str_ptr, ptr.as_ptr().byte_add(string_offset), str_len as usize);
        }

        // Ok, we can return the thing.
        VVRTVariant(NonNull::new_unchecked(ptr.as_ptr() as *mut _))
    }

    /// Create a new string from the content of the variant.
    #[no_mangle]
    unsafe extern "C" fn to_string(var: VVRTVariant) -> VVRTString {
        let discriminent = var.discriminent();
        let VariantDescription { ast_name, variant_name, arguments, .. } = &VARIANTS[discriminent];
        let mut pretty = format!("#({ast_name}.{variant_name}");

        arguments.iter().copied().enumerate().for_each(|(idx, arg)| {
            pretty.push_str(if idx == 0 { " " } else { ", " });
            pretty.push_str(&match arg {
                VVRTType::Number => Cow::Owned(format!("{}", var.get_number(idx as i32))),
                VVRTType::String => Cow::Borrowed(var.get_string()),
                ty => unreachable!("invalid variant {ast_name}.{variant_name} at idx {idx}, got type: {ty}"),
            });
        });

        pretty.push(')');
        VVRTString::from_vec(pretty.into())
    }

    #[no_mangle]
    unsafe extern "C" fn clone(this: VVRTVariant) -> VVRTVariant {
        let layout = this.get_value_layout(); // Get the layout of the value, because we need to
                                              // take into account the string part if any.
        let ptr = NonNull::new(alloc::alloc(layout) as *mut u64).expect("out of memory");
        std::ptr::copy_nonoverlapping(this.0.as_ptr() as *const u8, ptr.as_ptr() as *mut u8, layout.size());
        VVRTVariant(ptr)
    }

    /// Like [self::VVRTVariant_get] but for the string in the variant. Return default on type
    /// error. Here we don't have an index because we know that we have only one string in the
    /// variant and that it's at the last index. The string is allocated here and the codegen will
    /// need to insert a drop call...
    #[no_mangle]
    pub unsafe extern "C" fn get_string(this: VVRTVariant) -> VVRTString {
        Into::into(VVRTString::from_bytes(this.get_string().as_bytes()))
    }

    /// Like [self::VVRTVariant_get] but for i32. Return default on type error.
    #[no_mangle]
    pub unsafe extern "C" fn get_i32(this: VVRTVariant, idx: i32) -> i32 {
        this.get_number(idx).into()
    }

    /// Like [self::VVRTVariant_get] but for f32. Return default on type error.
    #[no_mangle]
    pub unsafe extern "C" fn get_f32(this: VVRTVariant, idx: i32) -> f32 {
        this.get_number(idx).into()
    }

    /// Clone the result, may be freed by the user.
    #[no_mangle]
    pub unsafe extern "C" fn get(this: VVRTVariant, idx: i32) -> VVRTAny {
        match this.offset_of_index(idx) {
            // For numbers we use directly the any to store them. Because of the discriminent
            // everything is already aligned in a good way.
            (ptr, VVRTType::Number) => *(ptr as *const VVRTAny),

            // We have a string, the first 4 bytes are the length of said string, the next things
            // are the content of said string.
            (ptr, VVRTType::String) => Into::into(VVRTString::from_bytes(slice::from_raw_parts(
                ptr.byte_add(size_of::<i32>()),
                *(ptr as *const i32) as usize,
            ))),

            _ => unreachable!("variant was not valid"),
        }
    }

    #[no_mangle]
    pub unsafe extern "C" fn deep_clone(this: VVRTVariant) -> VVRTVariant {
        VVRTVariant_clone(this)
    }

    /// Creates a new string!
    #[no_mangle]
    unsafe extern "C" fn get_variant_name(ptr: VVRTVariant) -> VVRTString {
        VVRTString::from_bytes(VARIANTS[ptr.discriminent()].variant_name.as_bytes())
    }

    /// Creates a new string!
    #[no_mangle]
    unsafe extern "C" fn get_ast_name(ptr: VVRTVariant) -> VVRTString {
        VVRTString::from_bytes(VARIANTS[ptr.discriminent()].ast_name.as_bytes())
    }

    #[no_mangle]
    unsafe extern "C" fn drop(var @ VVRTVariant(ptr): VVRTVariant) {
        alloc::dealloc(ptr.as_ptr() as *mut _, var.get_value_layout());
    }

    #[rustfmt::skip] #[no_mangle] unsafe extern "C" fn eq(this: VVRTVariant, other: VVRTVariant) -> bool { PartialEq::eq(&this, &other) }
    #[rustfmt::skip] #[no_mangle] unsafe extern "C" fn ne(this: VVRTVariant, other: VVRTVariant) -> bool { PartialEq::ne(&this, &other) }
}
