#![forbid(unsafe_code)]

//! A small string, with short string optimization.

use serde::{Deserialize, Serialize};
use smol_str::SmolStr;
use std::{borrow::Borrow, fmt::Display, ops::Deref};

/// A string as used in [crate::tokenizer::TokenType] and such. Does short string optimization.
#[derive(Clone, Debug, Default, Eq, PartialEq, Deserialize, Serialize, Hash, PartialOrd, Ord)]
#[serde(transparent)]
pub struct ShortString(SmolStr);

impl ShortString {
    /// Creates a new ShortString from the given text.
    pub fn new<T: Into<String> + AsRef<str>>(text: T) -> Self
    where
        SmolStr: From<T>,
    {
        ShortString(SmolStr::from(text))
    }

    /// Constructs inline variant of [ShortString].
    ///
    /// Panics if `text.len() > 23`. (See [SmolStr::new_inline])
    #[inline]
    pub const fn new_inline(text: &str) -> Self {
        ShortString(SmolStr::new_inline(text))
    }

    /// Constructs [ShortString]. Never allocates nor panic.
    #[inline]
    pub const fn new_static(text: &'static str) -> Self {
        if text.len() <= 23 {
            Self::new_inline(text)
        } else {
            ShortString(SmolStr::new_static(text))
        }
    }

    /// Returns a `&str` representation of the ShortString.
    pub fn as_str(&self) -> &str {
        self.0.as_str()
    }

    /// Returns the length of the ShortString.
    pub fn len(&self) -> usize {
        self.0.len()
    }

    /// Returns whether or not the ShortString is empty.
    pub fn is_empty(&self) -> bool {
        self.0.is_empty()
    }
}

impl Display for ShortString {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        self.0.fmt(f)
    }
}

impl Deref for ShortString {
    type Target = str;

    fn deref(&self) -> &Self::Target {
        &self.0
    }
}

impl Borrow<str> for ShortString {
    fn borrow(&self) -> &str {
        self.as_str()
    }
}

impl AsRef<str> for ShortString {
    fn as_ref(&self) -> &str {
        self.as_str()
    }
}

impl<T: Into<String> + AsRef<str>> From<T> for ShortString
where
    SmolStr: From<T>,
{
    fn from(value: T) -> Self {
        ShortString(SmolStr::from(value))
    }
}

impl From<&ShortString> for ShortString {
    fn from(value: &ShortString) -> Self {
        value.clone()
    }
}

impl From<ShortString> for String {
    fn from(value: ShortString) -> Self {
        value.0.to_string()
    }
}

impl FromIterator<char> for ShortString {
    fn from_iter<I: IntoIterator<Item = char>>(iter: I) -> Self {
        ShortString(SmolStr::from_iter(iter))
    }
}

#[test]
fn test_build_inline() {
    assert_eq!(ShortString::new_inline("12345678901234567890123").len(), 23);
    assert_eq!(ShortString::new_static("123456789012345678901230").len(), 24);
}

#[test]
#[should_panic]
fn test_oversized_inline() {
    let _ = ShortString::new_inline("123456789012345678901230");
}
