//! Utilities to extend the [XDGFolder] thing with config files: take into account the
//! deserialization, merge of configs and others.

use crate::{MaybeFolderList, XDGConfigFileSerde, XDGError, XDGFindBehaviour, XDGFolder};
use std::{fs, marker};

/// Search configurations in all config folders and merge them.
#[derive(Debug)]
#[non_exhaustive]
pub struct XDGConfigMerged;

/// Only get the config file with higher priority, which means searching in order:
/// 1. $XDG_CONFIG_HOME <- which has a default value.
/// 2. $XDG_CONFIG_DIRS <- which has a default value.
#[derive(Debug)]
#[non_exhaustive]
pub struct XDGConfigFirst;

/// Search configurations in all config folders and merge them. If the parsing for a configuration
/// file failed then log it and continue execution like normal, i.e. fail silently.
#[derive(Debug)]
#[non_exhaustive]
pub struct XDGConfigMergedSilent;

/// We will write one impl block for merged searches.
mod private_merged {
    use super::*;

    /// Search with merging.
    pub trait Sealed: Sized {
        fn try_read<Format: XDGConfigFileSerde>(config: &XDGConfig<Format, Self>) -> Result<Format, XDGError>;
    }

    impl Sealed for XDGConfigMergedSilent {
        fn try_read<Format: XDGConfigFileSerde>(config: &XDGConfig<Format, Self>) -> Result<Format, XDGError> {
            let mut result = config.search_files()?.into_iter().filter_map(|file| {
                let file = fs::read_to_string(file)
                    .map_err(|e| log::error!(target: "xdg", "{}", XDGError::ConfigIO(config.app.to_string(), config.get_file().to_string(), e)))
                    .ok()?;
                Format::deserialize(&file)
                    .map_err(|e| log::error!(target: "xdg", "deserialize error for {} on file {}: {e}", config.app, config.get_file()))
                    .ok()
            });

            let mut first = result
                .next()
                .ok_or_else(|| XDGError::ConfigNotFound(config.app.to_string(), config.get_file().to_string()))?;
            first.extend(result);
            Ok(first)
        }
    }

    impl Sealed for XDGConfigMerged {
        fn try_read<Format: XDGConfigFileSerde>(config: &XDGConfig<Format, Self>) -> Result<Format, XDGError> {
            let mut result = Iterator::collect::<Result<Vec<_>, _>>(config.search_files()?.into_iter().map(|file| {
                let file = fs::read_to_string(file)
                    .map_err(|err| XDGError::ConfigIO(config.app.to_string(), config.get_file().to_string(), err))?;
                Format::deserialize(&file).map_err(|err| {
                    log::error!(target: "xdg", "deserialize error for application {} on file {}: {err}", config.app, config.get_file());
                    XDGError::DeserializeError(config.app.to_string(), config.get_file().to_string())
                })
            }))?.into_iter();

            let mut first = result
                .next()
                .ok_or_else(|| XDGError::ConfigNotFound(config.app.to_string(), config.get_file().to_string()))?;
            first.extend(result);
            Ok(first)
        }
    }

    impl Sealed for XDGConfigFirst {
        fn try_read<Format: XDGConfigFileSerde>(config: &XDGConfig<Format, Self>) -> Result<Format, XDGError> {
            let file = config
                .search_files()?
                .into_first()
                .ok_or_else(|| XDGError::ConfigNotFound(config.app.to_string(), config.get_file().to_string()))?;
            let file = fs::read_to_string(file)
                .map_err(|err| XDGError::ConfigIO(config.app.to_string(), config.get_file().to_string(), err))?;
            Format::deserialize(&file)
            .map_err(|err| {
                log::error!(target: "xdg", "deserialize error on application {} on file {}: {err}", config.app, config.get_file());
                XDGError::DeserializeError(config.app.to_string(), config.get_file().to_string())
            })
        }
    }
}

/// A struct to contain informations about the app we are querying the config for. The behaviour of
/// the search can be changed.
/// ---
/// On the `Format` generic parameter:
/// - If the `Format` parameter implements [Serialize], then you can write or provide a default
///   that will be written to disk (or fail silently...)
/// - If the `Format` parameter implements [Serialize] and [Default], you can depend on the
///   default value if an error occured in the deserialization.
/// ---
/// On the `Merged` generic parameter:
/// - If the `Merged` parameter is [XDGConfigFirst], then only the file with the higher priority
///   is parsed.
/// - If the `Merged` parameter is [XDGConfigMerged] and `Format` implements [Extend] on itself,
///   then all the found files are parsed and then merged, an error is returned on the first
///   failure of the deserialization process.
#[derive(Debug)]
pub struct XDGConfig<'a, Format: XDGConfigFileSerde, Merged: private_merged::Sealed> {
    /// The application name.
    app: &'a str,

    /// The file name. If not present will be defaulted to `config` at resolution time.
    file: Option<&'a str>,

    /// Stores the format for deserialization.
    _type: marker::PhantomData<Format>,

    /// Stores the format for deserialization.
    _merged: marker::PhantomData<Merged>,
}

impl<'a, Format: XDGConfigFileSerde, Merged: private_merged::Sealed> XDGConfig<'a, Format, Merged> {
    /// By default we say that the default config file is ${FOLDER}/${APP}/config like many
    /// applications do.
    const DEFAULT_FILE: &'static str = "config";

    /// Create a new [XDGConfig] helper.
    #[inline]
    pub fn new(app: &'a str) -> Self {
        Self { app, file: Default::default(), _type: marker::PhantomData, _merged: marker::PhantomData }
    }

    /// Change the file to resolve.
    #[inline]
    pub fn file(self, file: &'a str) -> Self {
        Self { file: Some(file), ..self }
    }

    /// Get the name of the config file that we will try to read. Returns the default if not set by
    /// the user.
    #[inline]
    pub fn get_file(&self) -> &str {
        match self.file.as_ref() {
            Some(file) => file,
            None => Self::DEFAULT_FILE,
        }
    }

    /// Search all config files in all locations...
    #[inline]
    fn search_files(&self) -> Result<MaybeFolderList, XDGError> {
        XDGFolder::ConfigDirs.find(self.app, self.get_file(), XDGFindBehaviour::ExistingOnly)
    }

    /// Prepare config folders.
    #[inline]
    pub fn prepare_folder(&self) -> impl IntoIterator<Item = <MaybeFolderList as IntoIterator>::Item> {
        XDGFolder::ConfigDirs.prepare_folder()
    }

    /// When trying to read or write the default, we write the file with the same logic as the
    /// [XDGConfigFirst] variant, we add this function to reduce the code to write for the write
    /// logic...
    #[inline]
    fn to_config_first(&self) -> XDGConfig<Format, XDGConfigFirst> {
        let Self { app, file, .. } = *self;
        XDGConfig::<Format, XDGConfigFirst> { app, file, _type: marker::PhantomData, _merged: marker::PhantomData }
    }

    /// Try to read the config files and deserialize them. If one config file failed to be
    /// deserialized then returns an error if the merged was not silent, or just ignore the file if
    /// the merge was silent. If no config file where found this is an error. All the files are
    /// merged, this operation can't fail, implementations must fail silently or have sane defaults
    /// and prefer files with higher priority.
    #[inline]
    pub fn try_read(&self) -> Result<Format, XDGError> {
        Merged::try_read(self)
    }

    /// Try to read the config files and deserialize them. If an error is encountred at any point,
    /// log it and return the provided default if the merge was not silent, skip the file if it was
    /// silent. If needed the default value is written on the disk, note that this operation may
    /// fail silently.
    #[inline]
    pub fn read_or(&self, default: Format) -> Format {
        self.try_read().unwrap_or_else(|err| {
            log::error!(target: "xdg", "read error, return default value to user: {err}");
            self.to_config_first().write_silent(default)
        })
    }

    /// Try to read the config files and deserialize them. If an error is encountred at any point,
    /// log it and return the provided default if the merge was not silent, skip the file if it was
    /// silent. If needed the default value is written on the disk, note that this operation may
    /// fail silently.
    #[inline]
    pub fn read_or_else(&self, cb: impl FnOnce() -> Format) -> Format {
        self.try_read().unwrap_or_else(|err| {
            log::error!(target: "xdg", "read error, return default value to user: {err}");
            self.to_config_first().write_silent(cb())
        })
    }
}

impl<Format: XDGConfigFileSerde + Default, Merged: private_merged::Sealed> XDGConfig<'_, Format, Merged> {
    /// Try to read the config files and deserialize them. If an error is encountred at any point,
    /// log it and return the provided default if the merge was not silent, skip the file if it was
    /// silent. If needed the default value is written on the disk, note that this operation may
    /// fail silently.
    #[inline]
    pub fn read_or_default(&self) -> Format {
        self.try_read().unwrap_or_else(|err| {
            log::error!(target: "xdg", "read error, return default value to user: {err}");
            self.to_config_first().write_silent(Default::default())
        })
    }
}

impl<Format: XDGConfigFileSerde> XDGConfig<'_, Format, XDGConfigFirst> {
    /// Write a value to the default config file location, the one with the higher priority
    /// (usually the user's one)
    fn write(&self, value: &Format) -> Result<(), XDGError> {
        let content = Format::serialize(value)
            .map_err(|err| {
                log::error!(target: "xdg", "serialize error for application {} on file {}: {err}", self.app, self.get_file());
                XDGError::SerializeError(self.app.to_string(), self.get_file().to_string())
            })?;

        let path = XDGFolder::ConfigDirs
            .find(self.app, self.get_file(), XDGFindBehaviour::FirstOrCreate)?
            .into_first()
            .expect("the user must have at least one location to place a config file, permission or fs quota problem?");

        match fs::create_dir_all(
            path.parent()
                .ok_or(XDGError::ConfigFileHasNoParentFolder(self.app.to_string(), self.get_file().to_string()))?,
        ) {
            Err(err) if !matches!(err.kind(), std::io::ErrorKind::AlreadyExists) => {
                return Err(XDGError::ConfigIO(self.app.to_string(), self.get_file().to_string(), err))
            }
            _ => {}
        }

        fs::write(path, content)
            .map_err(|err| XDGError::ConfigIO(self.app.to_string(), self.get_file().to_string(), err))
    }

    /// Same as [XDGConfig::write] but log any error and fail silently, returning the value that
    /// was attempted to be written to disk.
    #[inline]
    fn write_silent(&self, value: Format) -> Format {
        if let Err(err) = self.write(&value) {
            log::error!(target: "xdg", "failed to write default with err: {err}")
        }
        value
    }
}
