use crate::*;
use std::{
    collections::hash_map::DefaultHasher,
    fs,
    hash::{Hash, Hasher},
    io::{Error as IoError, ErrorKind as IoErrorKind},
    path::Path,
};

/// An error can be comming from the IO thing or the resource could be already locked.
#[derive(Debug, thiserror::Error)]
pub enum LockError {
    #[error("{0}")]
    IO(#[from] IoError),

    #[error("resource is already locked")]
    AlreadyLocked,

    #[error("{0}")]
    XDG(#[from] XDGError),
}

/// Result type for lock functions.
pub type LockResult<T> = Result<T, LockError>;

/// Structure used to hold the lock file for a resource that can be hashed.
pub struct LockFile {
    path: PathBuf,
}

impl LockFile {
    #[inline]
    fn handle_xdg_error(err: XDGError) -> LockError {
        match err {
            XDGError::ConfigIO(_, _, err) => Self::handle_io_error(err),
            err => LockError::XDG(err),
        }
    }

    #[inline]
    fn handle_io_error(err: IoError) -> LockError {
        match err.kind() {
            IoErrorKind::AlreadyExists => LockError::AlreadyLocked,
            _ => LockError::IO(err),
        }
    }

    #[inline]
    fn create(app: impl AsRef<str>, file: String) -> LockResult<Self> {
        match XDGFolder::RuntimeDir
            .find(app, file, XDGFindBehaviour::NonExistingOnly)
            .map_err(Self::handle_xdg_error)?
        {
            MaybeFolderList::Empty => unreachable!("failed to find the runtime dir..."),
            MaybeFolderList::Folder(path) | MaybeFolderList::Many(path, _) => {
                log::debug!("create lock file at: {}", path.display());
                let _ = fs::OpenOptions::new()
                    .create_new(true)
                    .open(&path)
                    .map_err(Self::handle_io_error)?;
                Ok(Self { path })
            }
        }
    }

    /// Create a lock file for another file.
    #[inline]
    pub fn for_file(app: impl AsRef<str>, path: impl AsRef<Path>) -> LockResult<Self> {
        let mut s = DefaultHasher::new();
        path.as_ref().hash(&mut s);
        Self::create(app, format!("{:#X}.path.lock", s.finish()))
    }

    /// Create a lock file for an hashable ressource.
    #[inline]
    pub fn for_resource(app: impl AsRef<str>, rsc: impl Hash) -> LockResult<Self> {
        let mut s = DefaultHasher::new();
        rsc.hash(&mut s);
        Self::create(app, format!("{:#X}.rsc.lock", s.finish()))
    }

    /// Create a lock file for an ID.
    #[inline]
    pub fn for_id(app: impl AsRef<str>, id: u64) -> LockResult<Self> {
        Self::create(app, format!("{id:#X}.id.lock"))
    }
}

impl Drop for LockFile {
    #[inline]
    fn drop(&mut self) {
        match fs::remove_file(&self.path) {
            Ok(_) => log::debug!("remove lock file: {}", self.path.display()),
            Err(err) => log::error!("failed to remove lock file '{}': {err}", self.path.display()),
        }
    }
}
