use crate::*;
use std::{
    fs::{self, File, OpenOptions},
    io::{Error as IoError, ErrorKind as IoErrorKind},
    ops::{Deref, DerefMut},
    path::{Path, PathBuf},
};
use vvs_utils::rand;

/// An error can be comming from the IO thing or we have a collision with the temp file name.
#[derive(Debug, thiserror::Error)]
pub enum TempError {
    #[error("{0}")]
    IO(#[from] IoError),

    #[error("generated a duplicate temporary file")]
    Duplicate,

    #[error("{0}")]
    XDG(#[from] XDGError),

    #[error("failed to read env variable: {0}")]
    NoEnvVar(&'static str),
}

/// Result type for lock functions.
pub type TempResult<T> = Result<T, TempError>;

/// The temporary file structure.
pub struct TempFile {
    file: File,
    path: PathBuf,
    leak: bool,
}

impl Deref for TempFile {
    type Target = File;

    #[inline]
    fn deref(&self) -> &Self::Target {
        &self.file
    }
}

impl DerefMut for TempFile {
    #[inline]
    fn deref_mut(&mut self) -> &mut Self::Target {
        &mut self.file
    }
}

impl AsRef<File> for TempFile {
    fn as_ref(&self) -> &File {
        self.deref()
    }
}

impl AsMut<File> for TempFile {
    fn as_mut(&mut self) -> &mut File {
        self.deref_mut()
    }
}

impl Drop for TempFile {
    fn drop(&mut self) {
        if !self.leak {
            if let Err(err) = fs::remove_file(&self.path) {
                log::error!("failed to delete temp file '{}': {err}", self.path.display())
            }
        } else {
            log::debug!("leak temp file '{}'", self.path.display());
        }
    }
}

impl TempFile {
    #[inline]
    fn handle_xdg_error(err: XDGError) -> TempError {
        match err {
            XDGError::ConfigIO(_, _, err) => Self::handle_io_error(err),
            err => TempError::XDG(err),
        }
    }

    #[inline]
    fn handle_io_error(err: IoError) -> TempError {
        match err.kind() {
            IoErrorKind::AlreadyExists => TempError::Duplicate,
            _ => TempError::IO(err),
        }
    }

    #[cfg(unix)]
    #[inline]
    fn create(app: impl AsRef<str>, file: impl AsRef<str>) -> TempResult<PathBuf> {
        let list = XDGFolder::RuntimeDir
            .find(app, file, XDGFindBehaviour::NonExistingOnly)
            .map_err(Self::handle_xdg_error)?;
        match list {
            MaybeFolderList::Empty => todo!(),
            MaybeFolderList::Folder(path) | MaybeFolderList::Many(path, _) => Ok(path),
        }
    }

    #[cfg(not(unix))]
    fn create(app: impl AsRef<str>, file: impl AsRef<str>) -> TempResult<PathBuf> {
        use std::env;
        let folder = env::var("TEMP")
            .map(PathBuf::from)
            .or_else(|_| {
                env::var("TMP").map(PathBuf::from).or_else(|_| {
                    env::var("USERPROFILE").map(|profile| PathBuf::from(profile).join("AppData/Local/Temp"))
                })
            })
            .map_err(|_| TempError::NoEnvVar("TEMP, TMP, USERPROFILE"))?;
        if !folder.is_dir() {
            Err(IoError::new(
                IoErrorKind::NotFound,
                format!("folder doesn't exists or is not a directory: {}", folder.to_string_lossy()),
            )
            .into())
        } else {
            let file = "toto";
            Ok(folder)
        }
    }

    /// Create a new temporary file somewhere. Note that the creation pattern is predictible and
    /// not secure in any way. Moreover we depend on the fact that PID from dead processes are not
    /// reused between reboots, and that temp folder are cleared at reboots...
    #[inline]
    pub fn new(app: impl AsRef<str>) -> TempResult<Self> {
        let (pid, rand) = (std::process::id(), rand());
        let path = Self::create(app, format!("tmp.{pid}.{rand}"))?;
        log::debug!("create temporary file at: {}", path.display());
        let file = OpenOptions::new()
            .create_new(true)
            .write(true)
            .read(true)
            .open(&path)
            .map_err(Self::handle_io_error)?;
        Ok(Self { file, path, leak: false })
    }

    /// Allow to leak the file, it will still exists after the application finished to run.
    #[inline]
    pub fn leak(mut self) -> Self {
        self.leak = true;
        self
    }

    /// Get the path of the temp file.
    #[inline]
    pub fn path(&self) -> &Path {
        &self.path
    }
}
