//! Utility functions to follow the freedesktop specifications on config folders and such:
//! https://specifications.freedesktop.org/basedir-spec/basedir-spec-latest.html

pub mod file;

#[cfg(test)]
mod tests;

mod config;
mod folders;
mod paths;

pub use self::{config::*, folders::*, paths::*};

use std::{io::Error as IoError, path::PathBuf};

/// Get the user's home folder. Panics if the env variable was not found or it can't be
/// canonicalized.
#[inline]
pub fn home_folder() -> PathBuf {
    PathBuf::from(std::env::var("HOME").expect("failed to get the $HOME env variable"))
        .canonicalize()
        .expect("failed to canonicalize the $HOME folder path")
}

#[derive(Debug, thiserror::Error)]
pub enum XDGError {
    #[error("failed to find file {2} for application {1} with folder family {0}")]
    NotFound(XDGFolder, String, String),

    #[error("failed to find config file {1} for application {0}")]
    ConfigNotFound(String, String),

    #[error("failed to find parent folder of config file {1} for application {0}")]
    ConfigFileHasNoParentFolder(String, String),

    #[error("io error for config file {1} for application {0}: {2}")]
    ConfigIO(String, String, IoError),

    #[error("deserialization failed on file {1} for application {0}")]
    DeserializeError(String, String),

    #[error("serialization failed on file {1} for application {0}")]
    SerializeError(String, String),
}

/// Control the behaviour of the [XDGFolder::find] function on conflicts.
#[derive(Debug, Default, Clone, Copy, PartialEq, Eq)]
pub enum XDGFindBehaviour {
    /// Returns only the files that already exists. If no file is found, returns the file with the
    /// post priority that can be created, usually where the user want this file to exists. If you
    /// don't want to merge config/data files/folders this can be a sane default.
    #[default]
    FirstOrCreate,

    /// Returns only the files that already exists. If you want to merge config/data files/folders
    /// this can be a sane default.
    ExistingOnly,

    /// Only returns files that don't already exists and may be created. Note that even if a file
    /// can be created by the user, it won't necessarily means that you will be able to, keep in
    /// mind that race conditions may occur here.
    NonExistingOnly,

    /// Returns all files, the ones that exist and the ones that may be created. The user will need
    /// to handle the returned files as he sees fit.
    AllFiles,
}

/// We force the user to choose a way to serialize and deserialize the config file. This way, we
/// are format agnostique here and don't even depend on serde.
pub trait XDGConfigFileSerde: Sized + Extend<Self> {
    type Error: std::error::Error;

    /// Read the content of the config file.
    fn deserialize(input: &str) -> Result<Self, Self::Error>;

    /// Write the config file to string, so that we can write it latter into a file.
    fn serialize(&self) -> Result<String, Self::Error>;
}
