//! A PathBuf container as a sum type.

use std::{
    mem,
    path::{Path, PathBuf},
};

/// An enum to store maybe multiple folders, or none, with utility functions.
#[derive(Debug, Default, Clone, PartialEq, Eq)]
pub enum MaybeFolderList {
    /// No folder.
    #[default]
    Empty,

    /// Only one folder.
    Folder(PathBuf),

    /// Many folders.
    Many(PathBuf, Vec<PathBuf>),
}

/// An interator over [MaybeFolderList].
#[derive(Debug, Clone, PartialEq, Eq)]
pub struct MaybeFolderListIterator<'a> {
    next: Option<&'a PathBuf>,
    list: &'a [PathBuf],
}

impl MaybeFolderList {
    /// Is there no folder?
    #[inline]
    pub fn is_none(&self) -> bool {
        matches!(self, MaybeFolderList::Empty)
    }

    /// Is there folders?
    #[inline]
    pub fn is_some(&self) -> bool {
        !self.is_none()
    }

    /// Get the first folder, or the last folder. If no folder is present just returns [None].
    #[inline]
    pub fn first(&self) -> Option<&Path> {
        match self {
            MaybeFolderList::Empty => None,
            MaybeFolderList::Folder(f) | MaybeFolderList::Many(f, _) => Some(f.as_ref()),
        }
    }

    /// Append another [MaybeFolderList] at the end of a [MaybeFolderList]. Empties the other
    /// variable and try to reuse memory.
    pub fn append(&mut self, other: &mut MaybeFolderList) {
        match (self, other) {
            (this @ MaybeFolderList::Empty, other) => match mem::replace(other, MaybeFolderList::Empty) {
                MaybeFolderList::Empty => {}
                MaybeFolderList::Folder(f1) => *this = MaybeFolderList::Folder(f1),
                MaybeFolderList::Many(f1, fs) => *this = MaybeFolderList::Many(f1, fs),
            },

            (_, MaybeFolderList::Empty) => {}

            (MaybeFolderList::Many(_, fs), other @ MaybeFolderList::Folder(_)) => {
                let MaybeFolderList::Folder(f2) = mem::replace(other, MaybeFolderList::Empty) else {
                    unreachable!()
                };
                fs.push(f2);
            }
            (MaybeFolderList::Many(_, fs), other @ MaybeFolderList::Many(..)) => {
                let MaybeFolderList::Many(f2, ref mut fs2) = mem::replace(other, MaybeFolderList::Empty) else {
                    unreachable!()
                };
                fs.reserve(fs2.len() + 1);
                fs.push(f2);
                fs.append(fs2);
            }

            (this @ MaybeFolderList::Folder(_), other) => {
                let MaybeFolderList::Folder(f1) = mem::take(this) else {
                    unreachable!()
                };
                match mem::replace(other, MaybeFolderList::Empty) {
                    MaybeFolderList::Empty => unreachable!(),
                    MaybeFolderList::Folder(f2) => *this = MaybeFolderList::Many(f1, vec![f2]),
                    MaybeFolderList::Many(f2, mut fs) => {
                        fs.insert(0, f2);
                        *this = MaybeFolderList::Many(f1, fs);
                    }
                }
            }
        }
    }

    /// Get the first folder, or the last folder. If no folder is present just returns [None]. The
    /// difference from the [MaybeFolderList::first] function is that this function consumes the
    /// [MaybeFolderList].
    #[inline]
    pub fn into_first(self) -> Option<PathBuf> {
        match self {
            MaybeFolderList::Empty => None,
            MaybeFolderList::Folder(f) | MaybeFolderList::Many(f, _) => Some(f),
        }
    }

    /// Get an iterator over the contained folders.
    #[inline]
    pub fn iter(&self) -> MaybeFolderListIterator {
        let (next, list): (_, &[_]) = match self {
            MaybeFolderList::Empty => (None, &[]),
            MaybeFolderList::Folder(f) => (Some(f), &[]),
            MaybeFolderList::Many(f, fs) => (Some(f), &fs[..]),
        };
        MaybeFolderListIterator { next, list }
    }

    /// Create a list of folders from a description string and a separator.
    #[inline]
    pub(super) fn from_str_list(list: &str, separator: char) -> Self {
        let mut folders = list.split(separator);
        match folders.next() {
            None => MaybeFolderList::Empty,
            Some(first) => MaybeFolderList::Many(PathBuf::from(first), folders.map(PathBuf::from).collect()),
        }
    }
}

impl std::iter::Extend<PathBuf> for MaybeFolderList {
    fn extend<T: IntoIterator<Item = PathBuf>>(&mut self, iter: T) {
        match self {
            MaybeFolderList::Empty => {
                let mut iter = iter.into_iter();
                if let Some(next) = iter.next() {
                    *self = MaybeFolderList::Many(next, iter.collect());
                }
            }
            MaybeFolderList::Folder(f1) => *self = MaybeFolderList::Many(mem::take(f1), iter.into_iter().collect()),
            MaybeFolderList::Many(_, fs) => fs.extend(iter),
        }
    }
}

impl<'a> std::iter::Extend<&'a MaybeFolderList> for MaybeFolderList {
    fn extend<T: IntoIterator<Item = &'a MaybeFolderList>>(&mut self, iter: T) {
        let other = iter.into_iter().flat_map(MaybeFolderList::iter);
        match self {
            MaybeFolderList::Empty => *self = other.collect(),
            MaybeFolderList::Folder(f1) => *self = MaybeFolderList::Many(mem::take(f1), other.cloned().collect()),
            MaybeFolderList::Many(_, ref mut fs) => fs.extend(other.cloned()),
        }
    }
}

impl<S: AsRef<str>> From<S> for MaybeFolderList {
    #[inline]
    fn from(value: S) -> Self {
        MaybeFolderList::Folder(PathBuf::from(value.as_ref()))
    }
}

impl<P: AsRef<Path>> FromIterator<P> for MaybeFolderList {
    #[inline]
    fn from_iter<T: IntoIterator<Item = P>>(iter: T) -> Self {
        let mut folders = iter.into_iter();
        match folders.next() {
            None => MaybeFolderList::Empty,
            Some(first) => {
                MaybeFolderList::Many(first.as_ref().to_path_buf(), folders.map(|p| p.as_ref().to_path_buf()).collect())
            }
        }
    }
}

impl IntoIterator for MaybeFolderList {
    type Item = PathBuf;
    type IntoIter = <Vec<PathBuf> as IntoIterator>::IntoIter;

    #[inline]
    fn into_iter(self) -> Self::IntoIter {
        match self {
            MaybeFolderList::Empty => vec![].into_iter(),
            MaybeFolderList::Folder(f) => vec![f].into_iter(),
            MaybeFolderList::Many(f, mut fs) => {
                fs.insert(0, f);
                fs.into_iter()
            }
        }
    }
}

impl<'a> Iterator for MaybeFolderListIterator<'a> {
    type Item = &'a PathBuf;

    #[inline]
    fn next(&mut self) -> Option<Self::Item> {
        let next = self.next?;
        match self.list.split_first() {
            Some((next, list)) => {
                self.next = Some(next);
                self.list = list;
            }
            None => self.next = None,
        }
        Some(next)
    }
}
