#!/usr/bin/env cargo script

//! ```cargo
//! [package]
//! name = "topo"
//! version = "0.1.0"
//! edition = "2021"
//!
//! [dependencies]
//! hwlocality = { version = "1.0.0-alpha.1", features = ["hwloc-latest", "vendored"] }
//! anyhow = "*"
//! ```

use anyhow::{anyhow, Result};
use hwlocality::{
    cpu::binding::CpuBindingFlags,
    object::types::ObjectType,
    topology::support::{DiscoverySupport, FeatureSupport},
    Topology,
};
use std::os::unix::thread::JoinHandleExt;

fn main() -> Result<()> {
    let topology = Topology::new()?;
    if !topology.supports(FeatureSupport::discovery, DiscoverySupport::pu_count) {
        println!("This example needs accurate reporting of PU objects");
        return Ok(());
    }
    let Some(cpu_support) = topology.feature_support().cpu_binding() else {
        println!("This example requires CPU binding support");
        return Ok(());
    };
    if !(cpu_support.get_thread() && cpu_support.set_thread()) {
        println!("This example needs support for querying and setting thread CPU bindings");
        return Ok(());
    }

    let core_depth = topology.depth_or_below_for_type(ObjectType::Core)?;
    let mut handles: Vec<_> = Vec::new();

    for (idx, core) in topology.objects_at_depth(core_depth).enumerate() {
        let handle = std::thread::spawn(move || -> Result<()> {
            // Worker thread
            std::thread::sleep(std::time::Duration::from_secs_f64(3.0));
            Ok(())
        });
        let tid = handle.as_pthread_t();

        let mut bind_to = core
            .cpuset()
            .ok_or_else(|| anyhow!("CPU cores should have CpuSets"))?
            .clone_target();
        bind_to.singlify();

        let before = topology.thread_cpu_binding(tid, CpuBindingFlags::empty())?;
        topology.bind_thread_cpu(tid, &bind_to, CpuBindingFlags::empty())?;
        let after = topology.thread_cpu_binding(tid, CpuBindingFlags::empty())?;

        println!("- Thread {idx} binding: {before:?} -> {after:?}");
        handles.push(handle);
    }
    Ok(())
}

